import json

from datetime import datetime
from dateutil import rrule
from django import forms
from django.core.exceptions import ValidationError

from ambition_utils.rrule.constants import RecurrenceEnds
from ambition_utils.rrule.models import RRule

from ambition_utils.time_helpers import get_time_zones

FREQ_CHOICES = (
    (rrule.DAILY, 'Daily'),
    (rrule.WEEKLY, 'Weekly'),
    (rrule.MONTHLY, 'Monthly'),
    (rrule.YEARLY, 'Yearly'),
)


ENDS_CHOICES = (
    (RecurrenceEnds.NEVER, 'Never'),
    (RecurrenceEnds.AFTER, 'After'),
    (RecurrenceEnds.ON, 'On'),
)


REPEAT_BY_CHOICES = (
    ('DAY_OF_THE_MONTH', 'Day of the month'),
    ('DAY_OF_THE_WEEK_START', 'Day of the week counting from the beginning of the month'),
    ('DAY_OF_THE_WEEK_END', 'Day of the week counting backwards from the end of the month'),
    # ('FIRST_DAY_OF_MONTH', 'First day of Month'),
    ('DAY_OF_THE_MONTH_END', 'Last day of Month'),
)


class RRuleForm(forms.Form):
    """
    Handles submission of data for populating rrule objects. The field names are based on the rrule
    params defined here http://dateutil.readthedocs.io/en/stable/rrule.html
    """

    # Date from which the recurrence will be started from. This might not always be the first recurrence in the series
    dtstart = forms.DateField(
        error_messages={'required': 'Starts on is required'},
    )

    # The hour for each recurrence (0-23)
    byhour = forms.IntegerField(
        error_messages={'required': 'Hour is required'},
    )

    # The minute for each recurrence (0-59)
    byminute = forms.IntegerField(required=False)

    # The time zone which the submitted time is treated as. The submitted time is converted to utc
    time_zone = forms.ChoiceField(
        choices=get_time_zones(return_as_tuple=True),
        error_messages={'required': 'Time Zone is required'},
    )

    # Type of interval - daily, weekly, etc
    freq = forms.ChoiceField(
        choices=FREQ_CHOICES,
        error_messages={'required': 'Frequency is required'},
    )

    # How often does the event repeat? This is related to the repeats value.
    interval = forms.IntegerField()

    # Day checkboxes, required if frequency is weekly. json encoded list of day numbers ex: [0,2,4]
    byweekday = forms.CharField(required=False)

    # Required if frequency is monthly and day of week. json encoded list of nth day data ex: [[1,-2]]
    bynweekday = forms.CharField(required=False)

    # Only required to monthly frequency
    repeat_by = forms.ChoiceField(choices=REPEAT_BY_CHOICES, required=False)

    # Choice of how the recurrence can be ended
    ends = forms.ChoiceField(choices=ENDS_CHOICES)

    # Number of times the recurrence will occur. Only required if ends is set to AFTER
    count = forms.IntegerField(required=False)

    # Date when the recurrence will end. Only required if ends is set to ON. The 'until' date might not be the last
    # recurrence date
    until = forms.DateField(required=False)

    def clean_freq(self):
        """
        Make sure the frequency is an integer
        """
        return int(self.data.get('freq', -1))

    def clean(self):
        """
        Perform additional form validation based on submitted params
        """
        cleaned_data = super().clean()

        if self.errors:
            return cleaned_data

        # Check if count is required
        if self.cleaned_data['ends'] == RecurrenceEnds.AFTER and not self.cleaned_data['count']:
            raise ValidationError('Number of occurrences is required')

        # Check if until is required
        if self.cleaned_data['ends'] == RecurrenceEnds.ON and not self.cleaned_data['until']:
            raise ValidationError('Ending date is required')

        # Unset until date if end ON is not selected
        if self.cleaned_data.get('ends') != RecurrenceEnds.ON:
            self.cleaned_data['until'] = ''

        # Check end date is after start date
        if self.cleaned_data.get('until') and self.cleaned_data.get('until') <= self.cleaned_data.get('dtstart'):
            raise ValidationError('End date must be after the start date')

        # Check if byweekday is required
        if self.cleaned_data['freq'] == rrule.WEEKLY and not self.cleaned_data['byweekday']:
            raise ValidationError('At least one day choice is required')

        # Check if repeat_by is required
        if self.cleaned_data['freq'] == rrule.MONTHLY and not self.cleaned_data.get('repeat_by'):
            raise ValidationError('Repeat by is required')

        return cleaned_data

    def clean_byminute(self):
        """
        Set optional byminute value to 0 if not defined
        """

        if 'byminute' not in self.data:
            return 0
        else:
            return self.cleaned_data['byminute']

    def clean_byweekday(self):
        """
        Decode the byweekday option
        """
        value = []

        try:
            value = json.loads(self.data.get('byweekday', '[]'))
        except ValueError:
            pass

        return value

    def clean_bynweekday(self):
        """
        Decode the bynweekday option
        """
        value = []

        try:
            value = json.loads(self.data.get('bynweekday', '[]'))
        except ValueError:
            pass

        return value

    def get_rrule_params_from_cleaned_data(self):
        rrule_freq = self.cleaned_data['freq']
        start_date = self.cleaned_data['dtstart']

        # Convert date to datetime
        start_datetime = datetime.combine(start_date, datetime.min.time())

        # Build params for rrule object
        params = {
            'freq': rrule_freq,
            'dtstart': start_datetime,
            'byhour': self.cleaned_data.get('byhour'),
            'byminute': self.cleaned_data.get('byminute'),
            'interval': self.cleaned_data.get('interval', 1),
        }

        if self.cleaned_data.get('count'):
            params['count'] = self.cleaned_data.get('count')

        if self.cleaned_data.get('until'):
            # Convert date to datetime
            until = datetime.combine(self.cleaned_data.get('until'), datetime.min.time())

            # Add hour to until datetime if it exists
            until = until.replace(hour=params['byhour'], minute=params['byminute'])
            params['until'] = until

        # Add day choices
        if self.cleaned_data.get('freq') == rrule.WEEKLY:
            params['byweekday'] = self.cleaned_data.get('byweekday')

        # Add repeat by choices
        if self.cleaned_data.get('freq') == rrule.MONTHLY:
            if self.cleaned_data.get('repeat_by') == 'DAY_OF_THE_MONTH':
                params['bymonthday'] = start_datetime.day
            elif self.cleaned_data.get('repeat_by') == 'DAY_OF_THE_MONTH_END':
                params['bymonthday'] = -1
            else:
                params['byweekday'] = self.cleaned_data['bynweekday'][0][0]
                params['bysetpos'] = self.cleaned_data['bynweekday'][0][1]

        # Return the params
        return params


class RecurrenceForm(RRuleForm):
    """
    Converts rrule options into an rrule model
    """

    # Optional rrule object that we are modifying
    rrule = forms.ModelChoiceField(queryset=RRule.objects.all(), required=False)

    # Optional rrule exclusion params
    rrule_exclusion = forms.CharField(required=False)

    def __init__(self, *args, **kwargs):
        # Remove the instance param if it exists. Model forms will try to pass this, but it isn't used here and will
        # cause the base form init to fail
        kwargs.pop('instance', None)

        super(RecurrenceForm, self).__init__(*args, **kwargs)

    def clean_rrule_exclusion(self):
        # Get the value from data
        rrule_exclusion = self.cleaned_data.get('rrule_exclusion')
        if not rrule_exclusion:
            return None

        # Parse the exclusion options
        rrule_exclusion = json.loads(rrule_exclusion)
        rrule_exclusion['time_zone'] = self.cleaned_data['time_zone']

        # Validate the options
        rrule_exclusion_form = RRuleForm(data=rrule_exclusion)
        if not rrule_exclusion_form.is_valid():
            raise ValidationError('Exclusion options invalid')

        # Return the exclusion params
        return rrule_exclusion_form.get_rrule_params_from_cleaned_data()

    def save(self, **kwargs):
        """
        Saves the RRule model and returns it
        """
        # Keep track if this is an existing rrule that needs occurrence updated
        need_to_refresh_next_occurrence = False

        # Get the rrule model from the cleaned data
        rrule_model = self.cleaned_data.get('rrule')
        if rrule_model:
            # Refresh if the next occurrence is not expired.
            need_to_refresh_next_occurrence = rrule_model.next_occurrence > datetime.utcnow()
        else:
            # Use the recurrence passed into save kwargs
            rrule_model = kwargs.get('recurrence') or RRule()

        # Create or update the rule
        rrule_model.rrule_params = self.get_rrule_params_from_cleaned_data()
        rrule_model.rrule_exclusion_params = self.cleaned_data.get('rrule_exclusion')
        rrule_model.time_zone = self.cleaned_data.get('time_zone')
        for key, value in kwargs.items():
            if hasattr(rrule_model, key):
                # This try except is because some field names might be reverse foreign key relationships
                try:
                    setattr(rrule_model, key, value)
                except TypeError:  # pragma: no cover
                    pass

        # Check if this was an existing model that needs to have its next occurrence updated
        if need_to_refresh_next_occurrence:
            rrule_model.refresh_next_occurrence()

        rrule_model.save()

        # Return the rule
        return rrule_model



import datetime
import sys
from numbers import Number
from typing import Union, Iterable, List


from django.db import models
from django.utils.functional import cached_property
import fleming
from tdigest import TDigest

from manager_utils import ManagerUtilsManager, ManagerUtilsQuerySet


class BadPercentileValue(Exception):
    pass


class BadAnomalyType(Exception):
    pass


class AnomalyBaseQueryset(ManagerUtilsQuerySet):
    def unprocessed(self, assume_now=None):
        if assume_now is None:
            assume_now = datetime.datetime.utcnow()
        today = fleming.floor(assume_now, day=1)

        return self.filter(last_modified__lt=today)


class AnomalyBaseManager(ManagerUtilsManager):
    def get_queryset(self):
        return AnomalyBaseQueryset(self.model)

    def unprocessed(
            self,
            assume_now: Union[datetime.datetime, None] = None
    ) -> AnomalyBaseQueryset:
        return self.get_queryset().unprocessed(assume_now=assume_now)


class AnomalyBase(models.Model):
    """
    Override this class to create an Anomaly Detector.  Think of this as a key-value store
    for anomalies.  You update instances with observed data.  This teaches the detector what
    "normal" is.  You then check values against what it has learned using the .detect(...) method.

    You are free to add whatever fields you want to your detector as long as they don't
    collide with the field names on this base class.

    You must define a get_uid() method on your class to populate the uid field.
    """
    # set whether or not this detector will use the blob to do incremental updates
    # When inheriting from this class, you must specify whether or not this detector will be incremental.
    # Incremental detectors use the tdigest blob.  Non-incremental detectors rely on outside code for updates
    # (for example, direct sql computation)
    IS_INCREMENTAL = True
    PERCENTILE_LOW_DEFAULT = 5
    PERCENTILE_HIGH_DEFAULT = 95

    class Meta:
        abstract = True

    # unique id to identify this anomaly detector
    uid = models.CharField(max_length=256, unique=True)

    # JSON blob to persist the state of the tdigest
    blob = models.JSONField(default=dict, null=True)

    # User defined percentiles below and above which an anomaly will happen
    percentile_low = models.FloatField(null=True)
    percentile_high = models.FloatField(null=True)

    # These configure the accuracy/space tradeoff for the t-digest state
    delta = models.FloatField(default=.01, null=True)
    K = models.FloatField(default=25, null=True)

    # These are auto-populated and should not be altered by the user
    # I'm indexing these because we will likely want to do a fast search
    # for non-null values.
    threshold_low = models.FloatField(null=True, db_index=True)
    threshold_high = models.FloatField(null=True, db_index=True)

    # This holds the number of ingested values in the database.  Note that this is a duplicate of what
    # is in the digest for incremental detector.  You shouldn't access this field directly.  Use the
    # .count property instead
    num_values_ingested = models.IntegerField(default=0)

    # keeps track of the last time updated
    last_modified = models.DateTimeField(db_index=True)

    objects = AnomalyBaseManager()

    def compute_uid(self) -> str:
        """
        Override this method with code to populate your uid field
        """
        raise NotImplementedError('you must define a .compute_uid() method')

    @property
    def count(self) -> int:
        if self.IS_INCREMENTAL:
            self.num_values_ingested = self.digest.n
        return self.num_values_ingested

    @count.setter
    def count(self, new_count: int) -> None:
        if self.IS_INCREMENTAL:
            raise ValueError('Can\'t set count on an incrmental anomaly detector')
        else:
            self.num_values_ingested = new_count

    @cached_property
    def digest(self) -> TDigest:
        """
        This property pulls from the database to populate a tdigest instance
        """
        dig = TDigest(self.delta, self.K)
        if self.blob:
            dig.update_from_dict(self.blob)
        return dig

    @property
    def min_num_points_high(self) -> float:
        if self.percentile_high is None:
            return sys.maxsize
        return 100. / (100. - self.percentile_high)

    @property
    def min_num_points_low(self) -> float:
        if self.percentile_low is None:
            return sys.maxsize
        return 100. / self.percentile_low

    def update(self, data, reset_threshold=False):
        """
        This method can be overwritten.  It defaults to updating the tdigest.  If you overwrite
        this method, it is your responsibility to update the num_values_ingested field with
        the number of points you are using in the update.
        """
        return self._update_digest(data, reset_threshold)

    def _update_digest(
            self,
            data: Iterable[Union[float, int]],
            reset_thresholds: bool
    ) -> None:
        """
        Call this method with data observations.  These updates are incorporated into internal
        state that will be able to efficiently detect future abnormal values.

        :param data: either a number or an iterable of numbers
        :param reset_thresholds: Setting to True will force the thresholds to update.  This is a
                                pretty big efficiency hit, but you may want to update and check
                                data without hitting the database.  This lets you do that.
        """
        if not self.IS_INCREMENTAL:
            raise BadAnomalyType('update() can only be called for incremental anomaly detector')

        if hasattr(data, '__iter__'):
            self.digest.batch_update(data)
        else:
            self.digest.update(data)

        if reset_thresholds:
            self._set_thresholds_from_tdigest()

    def _detect_point(self, val: Number) -> int:
        """
        compares a single number with thresholds to determine if it is an anomaly
        :param val:  a number.
        :return: -1: low_anomaly,  0: no_anomaly, 1: high_anomaly
        """
        anomaly = 0
        if self.count > self.min_num_points_low and self.threshold_low is not None and val < self.threshold_low:
            anomaly = -1

        if self.count > self.min_num_points_high and self.threshold_high is not None and val > self.threshold_high:
            anomaly = 1
        return anomaly

    def detect(
            self,
            data: Union[Number, Iterable[Number]]
    ) -> Union[int, Iterable[int]]:
        """
        Checks whether data is an anomaly
        :param data: either a number or an iterable of numbers
        :return: either a number or list of numbers depending on input
                 -1 represents low anomaly
                 +1 represents high anomaly
                 0 represents no anomaly
        """
        # These are actually okay typwise, but mypy doesn't like it
        if hasattr(data, '__iter__'):
            data_it: Iterable[Number] = data  # type: ignore
        else:
            data_it: Iterable[Number] = [data]  # type: ignore

        out: List[int] = [self._detect_point(val) for val in data_it]

        if len(out) == 1:
            return out[0]
        else:
            return out

    def _check_percentiles(self) -> None:
        """
        Raises an error if percentile settings look hokey
        """
        # set undefined percentiles to default values
        if self.percentile_low is None:
            self.percentile_low = self.PERCENTILE_LOW_DEFAULT
        if self.percentile_high is None:
            self.percentile_high = self.PERCENTILE_HIGH_DEFAULT

        # make sure percentiles look right
        if not 0. < self.percentile_low < 50.:
            raise BadPercentileValue('(0. < low_percentile < 50) is False')
        if not 50. <= self.percentile_high < 100.:
            raise BadPercentileValue('(50. < high_percentile < 100) is False')

    def _set_thresholds_from_tdigest(self) -> None:
        """
        Automatically sets thresholds based on desired percentiles and values
        already seen.
        """
        # set the high threshold if enough points have been seen
        if self.count > 0:
            self.threshold_high = self.digest.percentile(self.percentile_high)

        # set the low threshold if enough points have been seen
        if self.count > 0:
            self.threshold_low = self.digest.percentile(self.percentile_low)

    def _set_last_modified(self) -> None:
        self.last_modified = datetime.datetime.utcnow()

    def pre_save_hooks(self) -> None:
        """
        Everything here needs to be run to ensure state of the model is properly
        persisted to the database.
        """
        # explicitly implement last_modified logic so that bulk updates/upserts work
        self._set_last_modified()

        # make sure uid is set
        if not self.uid:
            self.uid = self.compute_uid()

        # make sure the percentiles are in range
        self._check_percentiles()

        # if there are enough points, set the thresholds properly for incremental detector
        if self.IS_INCREMENTAL:
            self._set_thresholds_from_tdigest()

        # serialize the digest to the blob
        self.blob = self.digest.to_dict()

        # accessing count property makes sure counts are properly mirrored from digest
        self.count

    def save(self, *args, **kwargs) -> None:
        """
        Make sure the pre-save hooks are called on each save
        """
        self.pre_save_hooks()
        super(AnomalyBase, self).save(*args, **kwargs)



import multiprocessing

import time

from datetime import datetime
from django.test import TransactionTestCase
from django import db

from ambition_utils.postgres_lock.lock import PostgresLockContext, PostgresLockException


def acquire_postgres_lock(queue, process_name, lock_name, lock_timeout, value=None, sleep=0):  # pragma: no cover
    try:
        with PostgresLockContext(lock_name, value=value, timeout=lock_timeout):
            # Create the start time
            start_time = datetime.utcnow()

            # Sleep for the passed amount of time
            time.sleep(sleep)

            # Get the end time
            queue.put({
                process_name: {
                    'start_time': start_time,
                    'end_time': datetime.utcnow(),
                    'exception': None
                }
            })
    except PostgresLockException as e:
        # Put the exception in the queue
        queue.put({
            process_name: {
                'start_time': None,
                'end_time': None,
                'exception': e
            }
        })


class PostgresLockTests(TransactionTestCase):
    """
    Test the postgres lock
    """

    def test_value(self):
        """
        Test that we can pass a value
        """
        with PostgresLockContext('test', value='1') as lock_context:
            self.assertTrue(lock_context.lock['value'], '1')
        with PostgresLockContext('test', value='2') as lock_context:
            self.assertTrue(lock_context.lock['value'], '2')
            self.assertTrue(lock_context.lock['previous_value'], '1')

    def test_lock_context(self):
        # Create a queue for responses
        queue = multiprocessing.Queue()

        # Create the processes
        process_one = multiprocessing.Process(
            target=acquire_postgres_lock,
            kwargs={
                'queue': queue,
                'process_name': 'one',
                'lock_name': 'test',
                'lock_timeout': 60,
                'sleep': 5
            }
        )
        process_two = multiprocessing.Process(
            target=acquire_postgres_lock,
            kwargs={
                'queue': queue,
                'process_name': 'two',
                'lock_name': 'test',
                'lock_timeout': 60,
                'sleep': 5
            }
        )

        # close django's database connections to avoid multiprocess issue
        db.connections.close_all()

        # Start the processes
        process_one.start()
        time.sleep(1)
        process_two.start()

        # Wait for the processes to complete
        process_one.join()
        process_two.join()

        # Read the responses
        response = {}
        while not queue.empty():
            response.update(queue.get())

        # Assert that we acquired the lock in the correct order
        self.assertTrue(
            (response['two']['start_time'] - response['one']['start_time']).total_seconds() >= 5
        )

    def test_lock_context_session_timeout(self):
        # Create a queue for responses
        queue = multiprocessing.Queue()

        # Create the processes
        process_one = multiprocessing.Process(
            target=acquire_postgres_lock,
            kwargs={
                'queue': queue,
                'process_name': 'one',
                'lock_name': 'test',
                'lock_timeout': 1,
                'sleep': 5
            }
        )
        process_two = multiprocessing.Process(
            target=acquire_postgres_lock,
            kwargs={
                'queue': queue,
                'process_name': 'two',
                'lock_name': 'test',
                'lock_timeout': 1,
                'sleep': 5
            }
        )

        # close django's database connections to avoid multiprocess issue
        db.connections.close_all()

        # Start the processes
        process_one.start()
        time.sleep(1)
        process_two.start()

        # Wait for the processes to complete
        process_one.join()
        process_two.join()

        # Read the responses
        response = {}
        while not queue.empty():
            response.update(queue.get())

        # Assert that process two raised a proper timeout exception
        self.assertTrue(
            isinstance(response['two']['exception'], PostgresLockException)
        )



from django.apps import AppConfig


class AnomalyTestsConfig(AppConfig):
    name = 'ambition_utils.anomaly.tests'
    verbose_name = 'Ambition Anomaly Tests'
    label = 'ambition_utils_anomaly_tests'



from contextlib import contextmanager
from django.db import transaction

from ambition_utils.transaction import decorators


@contextmanager
def durable_reset_atomic_for_testing():  # pragma: no cover
    """
    A context manager for resetting the durable state for easier testing of a durable
    function inside of a transaction test case.
    This is only meant to be used for testing and never inside of production code.

    Any code wrapped in this will be tricked into thinking its not in a current
    transaction and will only keep track of new transactions created moving forward.
    This will allow us to bypass the initial transaction created by the test case and
    only care about those created from the actual code block
    """

    # Keep track of the original methods
    original_atomic_enter = transaction.Atomic.__enter__
    original_atomic_exit = transaction.Atomic.__exit__
    original_is_in_atomic_block = decorators._is_in_atomic_block
    original_db_may_have_uncommitted_work = decorators._db_may_have_uncommitted_work
    transactions = []

    # Create patch methods
    def patch_atomic_enter(*args, **kwargs):
        transactions.append((args, kwargs))
        return original_atomic_enter(*args, **kwargs)

    def patch_atomic_exit(*args, **kwargs):
        if len(transactions):
            transactions.pop()
        return original_atomic_exit(*args, **kwargs)

    def patch_is_in_atomic_block():
        if len(transactions):
            return original_is_in_atomic_block()
        return False

    def patch_db_may_have_uncommitted_work():
        return patch_is_in_atomic_block()

    # Patch the methods
    transaction.Atomic.__enter__ = patch_atomic_enter
    transaction.Atomic.__exit__ = patch_atomic_exit
    decorators._is_in_atomic_block = patch_is_in_atomic_block
    decorators._db_may_have_uncommitted_work = patch_db_may_have_uncommitted_work

    # Run the wrapped method
    yield True

    # Restore the patched methods
    transaction.Atomic.__enter__ = original_atomic_enter
    transaction.Atomic.__exit__ = original_atomic_exit
    decorators._is_in_atomic_block = original_is_in_atomic_block
    decorators._db_may_have_uncommitted_work = original_db_may_have_uncommitted_work


class DurableResetAtomicForTestingPatcher:  # pragma: no cover
    """
    A patcher for testing durable atomic blocks in a transaction test case
    """

    def __init__(self):
        self.patcher = None

    def start(self):
        self.patcher = durable_reset_atomic_for_testing()
        self.patcher.__enter__()

    def stop(self):
        if self.patcher:
            self.patcher.__exit__(None, None, None)



# Generated by Django 2.2.28 on 2022-06-23 16:49

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('rrule', '0002_auto_20190422_2037'),
    ]

    operations = [
        migrations.AddField(
            model_name='rrule',
            name='related_object_content_type',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to='contenttypes.ContentType'),
        ),
        migrations.AddField(
            model_name='rrule',
            name='related_object_handler_name',
            field=models.TextField(blank=True, default=None, null=True),
        ),
        migrations.AddField(
            model_name='rrule',
            name='related_object_id',
            field=models.IntegerField(db_index=True, null=True),
        ),
    ]



from __future__ import annotations
from datetime import datetime, timedelta
from dateutil import parser
from dateutil.rrule import rrule, rruleset
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.utils.module_loading import import_string
from fleming import fleming
from manager_utils import bulk_update
from ambition_utils.fields import TimeZoneField
from typing import List
import copy
import pytz
import logging

LOG = logging.getLogger(__name__)


class RRuleManager(models.Manager):
    """
    Custom manager for rrule objects
    """

    def update_next_occurrences(self, rrule_objects=None):
        if rrule_objects is None:
            return
        for rrule_object in rrule_objects:
            rrule_object.update_next_occurrence(save=False)

        bulk_update(self, rrule_objects, ['last_occurrence', 'next_occurrence'])

        return rrule_objects

    @transaction.atomic
    def handle_overdue(self, **kwargs):
        """
        Handles any overdue rrules
        :param kwargs: the old optional kwarg filters for specifying additional occurrence handler filters
        """
        self.process_occurrence_handler_paths(**kwargs)
        self.process_related_model_handlers()

    def process_occurrence_handler_paths(self, **kwargs):
        """
        This is the old style of processing overdue rrules
        """
        # Get instances of all overdue recurrence handler classes
        instances = self.overdue_handler_class_instances(**kwargs)

        # Build a list of rrules that get returned from the handler
        rrules = []
        for instance in instances:
            rrules.extend(instance.handle())

        # Bulk update the next occurrences
        RRule.objects.update_next_occurrences(rrule_objects=rrules)

    def process_related_model_handlers(self):
        # Get the rrule objects that are overdue and need to be handled
        rrule_objects = self.get_queryset().filter(
            next_occurrence__lte=datetime.utcnow(),
            related_object_handler_name__isnull=False,
            related_object_id__isnull=False,
        ).prefetch_related('related_object')

        rrules_to_advance = []
        for rrule_object in rrule_objects:
            if hasattr(rrule_object.related_object, rrule_object.related_object_handler_name):
                rrules_to_advance.append(
                    getattr(rrule_object.related_object, rrule_object.related_object_handler_name)(rrule_object)
                )

        rrules_to_advance = [rrule_to_advance for rrule_to_advance in rrules_to_advance if rrule_to_advance]

        # Bulk update the next occurrences
        rrules_to_advance = RRule.objects.update_next_occurrences(rrule_objects=rrules_to_advance)

        return rrules_to_advance

    def overdue_handler_class_instances(self, **kwargs):
        """
        Returns a set of instances for any handler with an old next_occurrence
        """

        # Get the rrule objects that are overdue and need to be handled
        rrule_objects = self.get_queryset().filter(
            next_occurrence__lte=datetime.utcnow(),
            **kwargs
        ).distinct(
            'occurrence_handler_path'
        )

        # Return instances of the handler classes
        handler_classes = [
            rrule_object.get_occurrence_handler_class_instance()
            for rrule_object in rrule_objects
        ]
        handler_classes = [handler_class for handler_class in handler_classes if handler_class]
        return handler_classes


class RRule(models.Model):
    """
    Model that will hold rrule details and generate recurrences to be handled by the supplied handler
    """

    # Params used to generate the rrule
    rrule_params = models.JSONField()

    # Optional params used to generate the rrule exclusion
    rrule_exclusion_params = models.JSONField(default=None, blank=True, null=True)

    # Any meta data associated with the object that created this rule
    meta_data = models.JSONField(default=dict)

    # The timezone all dates should be converted to
    time_zone = TimeZoneField(default='UTC')

    # The last occurrence date that was handled
    last_occurrence = models.DateTimeField(null=True, default=None)

    # The next occurrence date that should be handled
    next_occurrence = models.DateTimeField(null=True, default=None)

    # A python path to the handler class used to handle when a recurrence occurs for this rrule
    # The configuration class must extend ambition_utils.rrule.handler.OccurrenceHandler
    occurrence_handler_path = models.CharField(max_length=500, blank=False, null=False)

    # Generic relation back to object to explicitly call expiration methods
    related_object_id = models.IntegerField(null=True, db_index=True)
    related_object_content_type = models.ForeignKey(ContentType, null=True, on_delete=models.PROTECT)
    related_object = GenericForeignKey('related_object_content_type', 'related_object_id')

    # The name of the method to call on the related_object when the recurrence has expired
    related_object_handler_name = models.TextField(default=None, null=True, blank=True)

    # An optional number of days to offset occurrences by
    day_offset = models.SmallIntegerField(blank=True, null=True)

    # Custom object manager
    objects = RRuleManager()

    def get_time_zone_object(self):
        """
        Returns the time zone object from pytz
        """
        if self.time_zone is None:
            return pytz.utc

        # There is a test for this but it still doesn't hit this block
        if isinstance(self.time_zone, str):  # pragma: no cover
            return pytz.timezone(self.time_zone)

        return self.time_zone

    def get_occurrence_handler_class_instance(self):
        """
        Gets an instance of the occurrence handler class associated with this rrule
        :rtype: ambition_utils.rrule.handler.OccurrenceHandler
        :return: The instance
        """
        try:
            handler_class = import_string(self.occurrence_handler_path)()
            return handler_class
        except:
            return None

    def get_rrule_set(self):
        """
        Returns the rrule set that will combine the rrule and optional exclusion rrule
        """
        rrule_set = rruleset()
        rrule_set.rrule(self.get_rrule())
        rrule_exclusion = self.get_rrule_exclusion()
        if rrule_exclusion:
            rrule_set.exrule(rrule_exclusion)
        return rrule_set

    def get_rrule(self):
        """
        Builds the rrule object by restoring all the params.
        """
        return self.get_rrule_from_params(self.rrule_params)

    def get_rrule_exclusion(self):
        """
        Builds the rrule exclusion object by restoring all the params.
        :rtype: rrule
        """
        return self.get_rrule_from_params(self.rrule_exclusion_params)

    def get_rrule_from_params(self, params):
        """
        Creates an rrule object from a dict of rrule params. Returns None if no params exists.
        The dtstart param will be converted to local time if it is set.
        :rtype: rrule
        """
        # Check for none or empty
        if not params:
            return None

        # Create a deep copy because we will manipulate
        params = copy.deepcopy(params)

        # Convert next scheduled from utc back to time zone
        if params.get('dtstart') and not hasattr(params.get('dtstart'), 'date'):
            params['dtstart'] = parser.parse(params['dtstart'])

        # Convert until date from utc back to time zone
        if params.get('until') and not hasattr(params.get('until'), 'date'):
            params['until'] = parser.parse(params['until'])

        # Always cache
        params['cache'] = True

        # Return the rrule
        return rrule(**params)

    def get_next_occurrence(self, last_occurrence=None, calculate_offset=True, force=False):
        """
        Builds the rrule and returns the next date in the series or None of it is the end of the series
        :param last_occurrence: The last occurrence that was generated
        :param calculate_offset: Should the given offset be calculated?
        :param force: If the next occurrence is none, force the rrule to generate another
        :rtype: rrule or None
        """
        # Get the last occurrence
        last_occurrence = last_occurrence or self.last_occurrence or datetime.utcnow()

        # Get the rule set
        rule_set = self.get_rrule_set()

        # Convert to local time zone for getting next occurrence, otherwise time zones ahead of utc will return the same
        last_occurrence = fleming.convert_to_tz(last_occurrence, self.get_time_zone_object(), return_naive=True)

        # Un-offset the last occurrence to match the rule_set's dates for .after() before offsetting again later
        if calculate_offset:
            last_occurrence = self.offset(last_occurrence, reverse=True)

        # Generate the next occurrence
        next_occurrence = rule_set.after(last_occurrence)

        # If next occurrence is none and force is true, force the rrule to generate another date
        if next_occurrence is None and force:
            # Keep a reference to the original rrule_params
            original_rrule_params = {}
            original_rrule_params.update(self.rrule_params)

            # Remove any limiting params
            self.rrule_params.pop('count', None)
            self.rrule_params.pop('until', None)

            # Refetch the rule set
            rule_set = self.get_rrule_set()

            # Generate the next occurrence
            next_occurrence = rule_set.after(last_occurrence)

            # Restore the rrule params
            self.rrule_params = original_rrule_params

        # If there is a next occurrence, convert to utc
        if next_occurrence:
            next_occurrence = self.convert_to_utc(next_occurrence)

            if calculate_offset:
                next_occurrence = self.offset(next_occurrence)

        # Return the next occurrence
        return next_occurrence

    def update_next_occurrence(self, save=True):
        """
        Sets the next_occurrence property to the next time in the series and sets the last_occurrence property
        to the previous value of next_occurrence. If the save option is True, the model will be saved. The
        save flag is typically set to False when wanting to bulk update records after updating the values
        of many models.
        :param save: Flag to save the model after updating the schedule.
        :type save: bool
        """
        if not self.next_occurrence:
            return None

        # Only handle if the current date is >= next occurrence
        if datetime.utcnow() < self.next_occurrence:
            return False

        self.last_occurrence = self.next_occurrence

        self.next_occurrence = self.get_next_occurrence(self.last_occurrence)

        # Only save if the flag is true
        if save:
            self.save(update_fields=['last_occurrence', 'next_occurrence'])

    def convert_to_utc(self, dt):
        """
        Treats the datetime object as being in the timezone of self.timezone and then converts it to utc timezone.
        :type dt: datetime
        """
        # Add timezone info
        dt = fleming.attach_tz_if_none(dt, self.get_time_zone_object())

        # Convert to utc
        dt = fleming.convert_to_tz(dt, pytz.utc, return_naive=True)

        return dt

    def offset(self, dt, reverse=False) -> datetime:
        """
        Offsets a given datetime by the number of days specified by day_offset.
        :param dt: The datetime to offset by day_offset
        :param reverse: Reverse the offset calculation.
        :return dt:
        """
        # The offset gets multiplied by 1 or -1 depending on offset direction
        multiplier = -1 if reverse else 1

        # Timezone is considered when not reversing offset for comparisons in rrule.after().
        return fleming.add_timedelta(
            dt,
            timedelta(days=self.day_offset * multiplier),
            within_tz=self.time_zone if not reverse else None
        ) if self.day_offset else dt

    def refresh_next_occurrence(self, current_time=None):
        """
        Sets the next occurrence date based on the current rrule param definition. The date will be after the
        specified current_time or utcnow.
        :param current_time: Optional datetime object to compute the next time from
        """
        # Get the current time or go off the specified current time
        current_time = current_time or datetime.utcnow()

        # Next occurrence is in utc here
        next_occurrence = self.get_next_occurrence(last_occurrence=current_time)

        if next_occurrence:
            # Only set if the new time is still greater than now.
            # Offset date if applicable.
            if next_occurrence > datetime.utcnow():
                self.next_occurrence = self.offset(next_occurrence)
        else:
            self.next_occurrence = next_occurrence

    def pre_save_hooks(self):
        self.set_date_objects()

    def set_date_objects(self):
        """
        Ensure that all the date keys are properly set on all rrule params
        """
        # A cloned object will not have a primary key but does have a next_occurrence that should not be reset
        # to the first date in a series.
        is_new = self.pk is None and self.last_occurrence is None

        # Convert the rrule and exclusion rrule params to properly set date keys
        self.set_date_objects_for_params(self.rrule_params, is_new=is_new)
        self.set_date_objects_for_params(self.rrule_exclusion_params, is_new=is_new)

        # Check if this is a new rrule object
        if is_new:
            # Get the first scheduled time according to the rrule (this converts from utc back to local time)
            self.next_occurrence = self.get_rrule_set()[0]

            # Convert back to utc before saving
            self.next_occurrence = self.convert_to_utc(self.next_occurrence)

            # Offset, if applicable, for new objects.
            self.next_occurrence = self.offset(self.next_occurrence)

    def set_date_objects_for_params(self, params, is_new=False):
        """
        Give an rrule params object, ensure that the date keys are properly set and properly converted to strings
        """
        # Check for no params
        if not params:
            return params

        # Check if this is a new rrule object
        if is_new:
            # Convert next scheduled from utc back to time zone
            if params.get('dtstart') and not hasattr(params.get('dtstart'), 'date'):
                params['dtstart'] = parser.parse(params['dtstart'])

            # Convert until date from utc back to time zone
            if params.get('until') and not hasattr(params.get('until'), 'date'):
                params['until'] = parser.parse(params['until'])

        # Serialize the datetime objects if they exist
        if params.get('dtstart') and hasattr(params.get('dtstart'), 'date'):
            params['dtstart'] = params['dtstart'].strftime('%Y-%m-%d %H:%M:%S')

        if params.get('until') and hasattr(params.get('until'), 'date'):
            params['until'] = params['until'].strftime('%Y-%m-%d %H:%M:%S')

    def save(self, *args, **kwargs):
        """
        Saves the rrule model to the database. If this is a new object, the first next_scheduled time is
        determined and set. The `dtstart` and `until` objects will be safely encoded as strings if they are
        datetime objects.
        """

        # Run any pre save hooks
        self.pre_save_hooks()

        # Call the parent save method
        super().save(*args, **kwargs)

    def get_dates(self, num_dates=20, start_date=None) -> List[datetime]:
        """
        Return a list of datetime objects the recurrence will generate, after the start date (if defined).
        :param num_dates: The maximum number of dates to calculate. Will stop at passed start_date
        :param start_date: The optional start date to begin generating dates after
        :return: A list of datetime objects
        """

        # Assert that we have dates
        assert num_dates > 0

        # Ensure that pre save hooks have been run
        self.pre_save_hooks()

        # Generate the dates
        dates = []
        try:
            # Capture the rule's first date for use in RRule.after() in the loop.
            rule_set = self.get_rrule_set()

            # Evaluate if the first date should be retained.
            d = self.convert_to_utc(rule_set[0])
            if not start_date or d > start_date:
                dates.append(self.offset(d))

            # Continue evaluating and appending dates to satisfy desired number,
            # retaining date for evaluation in the next iteration.
            # The offset is ignored for this comparison and applied at appending.
            while len(dates) < num_dates:
                d = self.get_next_occurrence(last_occurrence=d, calculate_offset=False)
                if d:
                    if not start_date or d > start_date:
                        dates.append(self.offset(d))
                else:
                    break
        except Exception:  # pragma: no cover
            pass

        # Return the generated dates
        return dates

    def generate_dates(self, num_dates=20):
        """
        DEPRECATED. Replaced by get_dates.
        Return a list of the next num_dates datetimes of the recurrence.
        """
        LOG.warning('generate_dates has been replaced by get_dates and will be removed in version 3.x.')
        return self.get_dates(num_dates)

    def clone(self) -> RRule:
        """
        Creates a clone of itself.
        """

        # Clear id to force a new object.
        clone = copy.deepcopy(self)
        clone.id = None
        clone.save()
        return clone

    def clone_with_day_offset(self, day_offset: int) -> RRule:
        """
        Creates a clone of a passed RRule object with day_offset set.
        clone() is not called to ensure .id is not set before .save() so offset is applied.
        The clone's next_occurrence is set to the offset of this object.
        :param day_offset: The number of days to offset the clone's start date. Can be negative.
        """
        clone = copy.deepcopy(self)
        clone.id = None
        clone.day_offset = day_offset
        clone.next_occurrence = clone.offset(clone.next_occurrence)
        clone.save()
        return clone

    @classmethod
    def get_dates_from_params(cls, rrule_params, time_zone=None, num_dates=20, start_date=None):
        time_zone = time_zone or pytz.utc
        rule = cls(rrule_params=rrule_params, time_zone=time_zone)

        return rule.get_dates(num_dates=num_dates, start_date=start_date)

    @classmethod
    def generate_dates_from_params(cls, rrule_params, time_zone=None, num_dates=20):
        """
        DEPRECATED. Replaced by get_dates_from_params.
        """
        LOG.warning(
            'generate_dates_from_params has been replaced by get_dates_from_params and will be removed in version 3.x.'
        )
        return cls.get_dates_from_params(rrule_params, time_zone, num_dates)



import pytz
from timezone_field.choices import standard
from timezone_field.fields import TimeZoneField as BaseTimeZoneField


class CastOnAssignDescriptor:
    """
    A property descriptor which ensures that `field.to_python()` is called on _every_ assignment to the field.
    This used to be provided by the `django.db.models.subclassing.Creator` class, which in turn
    was used by the deprecated-in-Django-1.10 `SubfieldBase` class, hence the reimplementation here.
    Copied from https://stackoverflow.com/questions/
    39392343/how-do-i-make-a-custom-model-field-call-to-python-when-the-field-is-accessed-imm
    """

    def __init__(self, field):
        self.field = field

    def __get__(self, obj, type=None):
        if obj is None:  # pragma: no cover
            return self
        return obj.__dict__[self.field.name]

    def __set__(self, obj, value):
        obj.__dict__[self.field.name] = self.field.to_python(value)


class CastOnAssignFieldMixin:
    """
    Makes use of CastOnAssignDescriptor which will cast the field to its expected data type upon each assignment
    rather than only being cast when fetched from the db. This is the old behavior in django and is meant to be
    used for backwards compatibility for application that still expect this behavior.
    """

    def contribute_to_class(self, cls, name, virtual_only=False):
        """
        Cast to the correct value every
        """
        super().contribute_to_class(cls, name, virtual_only)
        setattr(cls, name, CastOnAssignDescriptor(self))


class TimeZoneField(CastOnAssignFieldMixin, BaseTimeZoneField):
    # Define default tzs on the class to support 4.2.x
    default_tzs = [pytz.timezone(tz) for tz in pytz.all_timezones]
    default_choices = standard(default_tzs)

    @staticmethod
    def get_default_pytz_tzs():
        """
        For timezone field v5+ this is how to override pytz defaults
        Override to use all time zones instead of common by default
        """
        return [pytz.timezone(tz) for tz in pytz.all_timezones]

    @classmethod
    def get_all_choices(cls):
        """
        Gets all timezones to use in a forms.ChoiceField for form validation
        """
        return [
            (tz, tz)
            for tz in pytz.all_timezones
        ]

    @classmethod
    def get_common_choices(cls):
        """
        Gets common timezones to use in a forms.ChoiceField for form validation
        """
        return [
            (tz, tz)
            for tz in pytz.common_timezones
        ]



from django.db import models


class PostgresLock(models.Model):
    """
    A model that will represent a postgres lock. A postgres lock is just a unique row
    on a model that we will insert or update a row to indicate that the row is locked.
    Postgres automatically locks the row and prevents others from writing until the transaction
    is complete. This allows us to treat this exactly like a locking system
    """

    # The name of the lock
    name = models.CharField(
        max_length=512,
        unique=True,
        primary_key=True,
    )

    # The last time the lock was updated
    time = models.DateTimeField(default=None)

    # An optional value to store with the lock
    value = models.TextField(default=None, null=True)

    # The previous value the last time the lock was acquired
    previous_value = models.TextField(default=None, null=True)



import datetime
import sys

from django.test import TestCase
from django.db import IntegrityError
from freezegun import freeze_time

from ambition_utils.anomaly.tests.models import IncrementalAnomaly, NonIncrementalAnomaly, FloatModel
from ambition_utils.anomaly.models import BadAnomalyType, BadPercentileValue


class NonIncrementalAnomalyTest(TestCase):
    def setUp(self):

        inserts = [FloatModel(value=nn) for nn in range(1, 101)]
        FloatModel.objects.bulk_create(inserts)

        NonIncrementalAnomaly.objects.create(name='non_incremental')

    def test_base_case(self):

        detector = NonIncrementalAnomaly.objects.get(name='non_incremental')
        detector.update()

        values = sorted(FloatModel.objects.all().values_list('value', flat=True))
        outliers = detector.detect(values)

        lows = {t[0] for t in zip(values, outliers) if t[1] == -1}
        highs = {t[0] for t in zip(values, outliers) if t[1] == 1}

        self.assertEqual(lows, {1, 2, 3, 4, 5})
        self.assertEqual(highs, {97, 98, 99, 100})


class IncrementalAnomalyTest(TestCase):
    def setUp(self):

        inserts = [FloatModel(value=nn) for nn in range(1, 101)]
        FloatModel.objects.bulk_create(inserts)

        IncrementalAnomaly.objects.create(name='incremental')

    def test_base_case(self):

        detector = IncrementalAnomaly.objects.get(name='incremental')

        values = FloatModel.objects.all().values_list('value', flat=True)
        detector.update(values)
        detector.save()

        outliers = detector.detect(values)

        lows = {t[0] for t in zip(values, outliers) if t[1] == -1}
        highs = {t[0] for t in zip(values, outliers) if t[1] == 1}

        self.assertEqual(lows, {1, 2, 3, 4, 5})
        self.assertEqual(highs, {96, 97, 98, 99, 100})


class IncrementalLimitedRangeTest(TestCase):
    def setUp(self):

        inserts = [FloatModel(value=nn % 4) for nn in range(1, 101)]
        FloatModel.objects.bulk_create(inserts)

        IncrementalAnomaly.objects.create(name='incremental')

    def test_base_case(self):
        detector = IncrementalAnomaly.objects.get(name='incremental')
        values = [nn % 4 for nn in range(1, 101)]
        detector.update(values)
        detector.save()

        self.assertEqual(detector.detect(list(range(-1, 6))), [-1, 0, 0, 0, 0, 1, 1])


class IncrementalRareEventTest(TestCase):
    def setUp(self):

        inserts = [FloatModel(value=nn % 4) for nn in range(1, 101)]
        # FloatModel.objects.bulk_create(inserts)
        for nn, insert in enumerate(inserts):
            if nn in [10, 11, 12]:
                inserts[nn].value = 5
            inserts[nn].save()

        IncrementalAnomaly.objects.create(name='incremental')
        self.inserts = inserts

    def test_base_case(self):
        detector = IncrementalAnomaly.objects.get(name='incremental')
        values = [m.value for m in self.inserts]
        for v in values:
            detector.update(v, reset_threshold=True)
        # detector.update(values)
        detector.save()
        anomalies = [detector.detect(v) for v in values]
        outliers = [t[0] for t in zip(values, anomalies) if t[1] != 0]
        self.assertEqual(outliers, [5, 5, 5])


class UnprocessedAnomalyTests(TestCase):
    def test_duplicate_uids_not_allowed(self):
        IncrementalAnomaly.objects.create(name='a')
        with self.assertRaises(IntegrityError):
            IncrementalAnomaly.objects.create(name='a')

    def test_get_unprocessed(self):
        names = ['a{}'.format(nn) for nn in range(5)]

        for nn, name in enumerate(names):
            detector = IncrementalAnomaly(name=name)
            if nn > 2:
                with freeze_time(datetime.datetime.utcnow() - datetime.timedelta(days=2)):
                    detector.save()
            else:
                detector.save()

        self.assertEqual({d.uid for d in IncrementalAnomaly.objects.unprocessed()}, {'a3', 'a4'})
        self.assertEqual(
            {
                d.uid for d in IncrementalAnomaly.objects.unprocessed(
                    assume_now=datetime.datetime.utcnow() + datetime.timedelta(days=2)
                )
            },
            {'a0', 'a1', 'a2', 'a3', 'a4'}
        )


class AnomalyUnitTests(TestCase):
    def test_zero_counts(self):
        inc_detector = IncrementalAnomaly.objects.get_or_create(name='non_incremental')[0]
        non_inc_detector = NonIncrementalAnomaly.objects.get_or_create(name='non_incremental')[0]

        self.assertEqual(inc_detector.count, 0)
        self.assertEqual(non_inc_detector.count, 0)

    def test_count_setting(self):
        inc_detector = IncrementalAnomaly.objects.get_or_create(name='non_incremental')[0]
        non_inc_detector = NonIncrementalAnomaly.objects.get_or_create(name='non_incremental')[0]

        non_inc_detector.count = 7
        non_inc_detector.save()
        self.assertEqual(NonIncrementalAnomaly.objects.get(name=non_inc_detector.uid).count, 7)

        with self.assertRaises(ValueError):
            inc_detector.count = 7

    def test_unassigned_percentiles(self):
        detector = NonIncrementalAnomaly(name='non_incremental')
        self.assertEqual(detector.min_num_points_high, sys.maxsize)
        self.assertEqual(detector.min_num_points_low, sys.maxsize)

    def test_bad_update(self):
        detector = IncrementalAnomaly(name='non_incremental')
        detector.IS_INCREMENTAL = False
        with self.assertRaises(BadAnomalyType):
            detector.update(7)

    def test_bad_percentiels(self):
        detector = IncrementalAnomaly(
            name='non_incremental',
            percentile_high=30,
            percentile_low=1
        )
        with self.assertRaises(BadPercentileValue):
            detector.save()

        detector = IncrementalAnomaly(
            name='non_incremental',
            percentile_high=90,
            percentile_low=60
        )
        with self.assertRaises(BadPercentileValue):
            detector.save()



# Generated by Django 2.2.28 on 2023-01-24 17:54

import ambition_utils.fields
from django.db import migrations
import ambition_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('ambition_utils_tests', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='fakemodel',
            name='cast_time_zone_field',
            field=ambition_utils.fields.TimeZoneField(default='utc', null=True),
        ),
        migrations.AddField(
            model_name='fakemodel',
            name='no_cast_time_zone_field',
            field=ambition_utils.fields.TimeZoneField(default='utc', null=True),
        ),
    ]



# Generated by Django 3.2.20 on 2023-08-02 15:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rrule', '0004_rrule_rrule_exclusion_params'),
    ]

    operations = [
        migrations.AddField(
            model_name='rrule',
            name='day_offset',
            field=models.SmallIntegerField(blank=True, null=True),
        ),
    ]



import tempfile
import os
from django.test import TestCase
from ambition_utils.tests.models import FakeModel
from ambition_utils.sql import StringSQL, FileSQL, queryset_to_sql


class TestSQL(TestCase):
    def setUp(self):
        self.simple_query = 'SELECT * FROM ambition_utils_tests_fakemodel;'
        self.param_query = 'SELECT * FROM ambition_utils_tests_fakemodel WHERE name=%(name)s;'
        self.context_query = 'SELECT * FROM {{table}};'
        self.insert_query = "INSERT INTO ambition_utils_tests_fakemodel (id, name) VALUES (DEFAULT, 'newname')"
        for nn in range(1, 4):
            FakeModel(name=f'n_{nn}').save()

    def test_tuples(self):
        sql = StringSQL(self.simple_query)
        tups = sql.to_tuples()
        self.assertEqual({t[1] for t in tups}, {'n_1', 'n_2', 'n_3'})

    def test_dataframe(self):
        sql = StringSQL(self.simple_query)
        sql.using_connection(sql._connection)
        df = sql.to_dataframe()
        self.assertEqual(self.simple_query, sql.raw_sql)
        self.assertEqual(set(df.name), {'n_1', 'n_2', 'n_3'})

    def test_no_return(self):
        sql = StringSQL(self.insert_query)
        df = sql.to_dataframe()
        self.assertTrue(df.empty)

    def test_params_named_tuples(self):
        sql = StringSQL(self.param_query)
        sql.with_params(dict(name='n_1'))
        tups = sql.to_named_tuples()
        self.assertEqual({t.name for t in tups}, {'n_1'})

    def test_context_dicts(self):
        sql = StringSQL(self.context_query)
        sql.with_context(dict(table='ambition_utils_tests_fakemodel'))
        tups = sql.to_dicts()
        self.assertEqual({t['name'] for t in tups}, {'n_1', 'n_2', 'n_3'})

    def test_abs_file_sql(self):
        with tempfile.NamedTemporaryFile('w') as query_file:
            query_file.write(self.simple_query)
            query_file.flush()

            sql = FileSQL(query_file.name, path_is_relative=False)
            df = sql.to_dataframe()
            self.assertEqual(set(df.name), {'n_1', 'n_2', 'n_3'})

            # Run query twice to use some cache hits
            df = sql.to_dataframe()
            self.assertEqual(set(df.name), {'n_1', 'n_2', 'n_3'})

    def test_rel_file_sql(self):
        with tempfile.NamedTemporaryFile('w') as query_file:
            query_file.write(self.simple_query)
            query_file.flush()

            rel_path = os.path.relpath(query_file.name, os.path.realpath(__file__))
            sql = FileSQL(rel_path, path_is_relative=True)
            df = sql.to_dataframe()
            self.assertEqual(set(df.name), {'n_1', 'n_2', 'n_3'})

    def test_queryset_to_sql(self):
        # Build a query
        qs = FakeModel.objects.order_by('name').values('name')

        # Get the postgres sql for that query
        query = queryset_to_sql(qs)

        # Run the postgres sql
        sql = StringSQL(query)
        sql_recs = sql.as_dicts()

        # Now run the acutal query
        qs_recs = list(qs)

        # Make sure they are the same
        self.assertEqual(
            [r['name'] for r in sql_recs],
            [r['name'] for r in qs_recs],
        )



# Generated by Django 2.2.28 on 2023-01-06 21:13

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('rrule', '0003_auto_20220623_1649'),
    ]

    operations = [
        migrations.CreateModel(
            name='Program',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.TextField()),
                ('start_called', models.IntegerField(default=0)),
                ('end_called', models.IntegerField(default=0)),
                ('end_recurrence', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='programs_end', to='rrule.RRule')),
                ('start_recurrence', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='programs_start', to='rrule.RRule')),
            ],
        ),
    ]



import datetime
import fleming

import pytz
from dateutil import rrule, parser
from django.test import TestCase
from django_dynamic_fixture import G
from freezegun import freeze_time

from ambition_utils.rrule.constants import RecurrenceEnds
from ambition_utils.rrule.forms import RecurrenceForm
from ambition_utils.rrule.handler import OccurrenceHandler
from ambition_utils.rrule.models import RRule
from ambition_utils.rrule.tests.models import Program


class MockHandler(OccurrenceHandler):
    """
    Mack handler for handling an occurrence during testing
    """

    def handle(self, rrule):
        return True


class HandlerOne(OccurrenceHandler):

    def handle(self):
        return RRule.objects.filter(
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne',
            next_occurrence__lte=datetime.datetime.utcnow(),
        ).order_by('id')


class HandlerTwo(OccurrenceHandler):

    def handle(self):
        return RRule.objects.filter(
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerTwo',
            next_occurrence__lte=datetime.datetime.utcnow(),
        ).order_by('id')


class HandlerThree(OccurrenceHandler):

    def handle(self):
        return RRule.objects.filter(
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerThree',
            next_occurrence__lte=datetime.datetime.utcnow(),
        ).order_by('id')


class RRuleManagerTest(TestCase):

    def test_update_next_occurrences(self):
        """
        Make sure that the correct occurrences are selected
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        rrule1 = G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne'
        )

        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        rrule2 = G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne'
        )

        self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 2))

        # Progress no rules
        with freeze_time('1-3-2017'):
            RRule.objects.update_next_occurrences()

        rrule1.refresh_from_db()
        rrule2.refresh_from_db()

        # Both should be progressed
        self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 2))

        # Progress all rules
        with freeze_time('1-3-2017'):
            RRule.objects.update_next_occurrences([rrule1, rrule2])

        rrule1.refresh_from_db()
        rrule2.refresh_from_db()

        # Both should be progressed
        self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 3))
        self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 3))

        # Progress specific rrules
        with freeze_time('1-4-2017'):
            RRule.objects.update_next_occurrences(rrule_objects=[rrule1])

        rrule1.refresh_from_db()
        rrule2.refresh_from_db()

        # One should be progressed
        self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 4))
        self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 3))

        # Make sure neither are progressed with passing an empty list
        with freeze_time('1-5-2017'):
            RRule.objects.update_next_occurrences(rrule_objects=[])

        rrule1.refresh_from_db()
        rrule2.refresh_from_db()

        # One should be progressed
        self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 4))
        self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 3))

    @freeze_time('1-1-2017')
    def test_run(self):
        """
        Should return the classes that have overdue rrule objects
        """

        # Make a program with HandlerOne that is not overdue
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne'
        )

        # Make an overdue program with HandlerOne
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne'
        )

        # Make an overdue program with HandlerTwo
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerTwo'
        )

        # Make a program with HandlerThree that is not overdue
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerThree'
        )

        # Get and verify the classes
        classes = {
            instance.__class__
            for instance in RRule.objects.overdue_handler_class_instances()
        }

        self.assertEqual(classes, {HandlerOne, HandlerTwo})

        # Handle overdue rrules
        RRule.objects.handle_overdue()

        # Check the recurrences. ids 2 and 3 should be updated
        recurrences = list(RRule.objects.order_by('id'))

        self.assertEqual(recurrences[0].next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(recurrences[1].next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(recurrences[2].next_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(recurrences[3].next_occurrence, datetime.datetime(2017, 1, 2))

        # For coverage, make handler 3 overdue
        with freeze_time('1-3-2017'):
            RRule.objects.handle_overdue()


class RRuleTest(TestCase):

    def test_related_object_handlers(self):
        """
        Verifies the behavior of rrule related object handlers
        """
        program = Program.objects.create(name='Program 1')
        start_rrule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2022, 6, 1, 9),
                'byhour': 9,
            },
            related_object=program,
            related_object_handler_name='handle_start_recurrence',
        )
        end_rrule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2022, 6, 1, 17),
                'byhour': 17,
            },
            related_object=program,
            related_object_handler_name='handle_end_recurrence',
        )
        program.start_recurrence = start_rrule
        program.end_recurrence = end_rrule
        program.save()

        # Make sure handlers are not called before date
        with freeze_time(datetime.datetime(2022, 5, 31)):
            RRule.objects.handle_overdue()
            program = Program.objects.get(id=program.id)
            self.assertEqual(program.start_called, 0)
            self.assertEqual(program.end_called, 0)
            self.assertEqual(program.start_recurrence.next_occurrence, datetime.datetime(2022, 6, 1, 9))
            self.assertEqual(program.end_recurrence.next_occurrence, datetime.datetime(2022, 6, 1, 17))

        # Make sure only start handler is called
        with freeze_time(datetime.datetime(2022, 6, 1, 9)):
            RRule.objects.handle_overdue()
            program = Program.objects.get(id=program.id)
            self.assertEqual(program.start_called, 1)
            self.assertEqual(program.end_called, 0)
            self.assertEqual(program.start_recurrence.next_occurrence, datetime.datetime(2022, 6, 2, 9))
            self.assertEqual(program.end_recurrence.next_occurrence, datetime.datetime(2022, 6, 1, 17))

        # Make sure only end handler is called
        with freeze_time(datetime.datetime(2022, 6, 1, 17)):
            RRule.objects.handle_overdue()
            program = Program.objects.get(id=program.id)
            self.assertEqual(program.start_called, 1)
            self.assertEqual(program.end_called, 1)
            self.assertEqual(program.start_recurrence.next_occurrence, datetime.datetime(2022, 6, 2, 9))
            self.assertEqual(program.end_recurrence.next_occurrence, datetime.datetime(2022, 6, 2, 17))

    def test_related_object_handlers_invalid_handler(self):
        """
        Hits the else block when the handler path is not valid
        """
        program = Program.objects.create(name='Program 1')
        start_rrule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2022, 6, 1, 9),
                'byhour': 9,
            },
            related_object=program,
            related_object_handler_name='handle_start_recurrence_fake',
        )
        end_rrule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2022, 6, 1, 17),
                'byhour': 17,
            },
            related_object=program,
            related_object_handler_name='handle_end_recurrence_fake',
        )
        program.start_recurrence = start_rrule
        program.end_recurrence = end_rrule
        program.save()

        # Make sure handlers are not called before date
        with freeze_time(datetime.datetime(2022, 7, 31)):
            RRule.objects.handle_overdue()

        # Occurrences should not be progressed
        program = Program.objects.get(id=program.id)
        self.assertEqual(program.start_recurrence.next_occurrence, datetime.datetime(2022, 6, 1, 9))
        self.assertEqual(program.end_recurrence.next_occurrence, datetime.datetime(2022, 6, 1, 17))

    def test_get_time_zone_object_none(self):
        """
        Should return utc when time zone is null
        """
        rrule = RRule()
        rrule.time_zone = None
        self.assertEqual(rrule.get_time_zone_object(), pytz.utc)

    def test_get_time_zone_object_str(self):
        """
        Should return an object when time zone is a string
        """
        rrule = RRule()
        rrule.time_zone = 'UTC'
        self.assertEqual(rrule.get_time_zone_object(), pytz.utc)

    def test_get_time_zone_object(self):
        """
        Should return the object when time zone is an object
        """
        rrule = RRule()
        rrule.time_zone = pytz.utc
        self.assertEqual(rrule.get_time_zone_object(), pytz.utc)

    def test_get_next_occurrence_first_is_occurrence(self):
        """
        First occurrence should be the dtstart
        """

        # Setup the params for creating the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
            'count': 3,
            'bymonthday': 1,
            'byhour': 10,
        }

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Assert initial values
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 1, 10))
        self.assertEqual(rule.next_occurrence, None)

        # For coverage run the handler
        with freeze_time('1-3-2017'):
            handler = MockHandler()

            self.assertTrue(handler.handle(None))

    def test_get_next_occurrence_first_is_not_occurrence(self):
        """
        First occurrence should be later than the dtstart
        """

        # Create the params to create the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
            'count': 3,
            'bymonthday': 1,
            'byhour': 10,
        }

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Assert the initial values
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 1, 10))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 4, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 4, 1, 10))
        self.assertEqual(rule.next_occurrence, None)

    def test_get_next_occurrence_force(self):
        """
        If there is no next occurrence but we want to know what it would have been
        """
        # Create the params to create the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
            'count': 1,
            'bymonthday': 1,
            'byhour': 10,
        }

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Assert the initial values
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 10))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 10))
        self.assertEqual(rule.next_occurrence, None)
        self.assertEqual(rule.get_next_occurrence(), None)
        self.assertEqual(rule.get_next_occurrence(force=True), datetime.datetime(2017, 3, 1, 10))

        # Coverage for returning early
        self.assertEqual(rule.update_next_occurrence(), None)

    def test_get_next_occurrence_dst(self):
        """
        What happens across DST changes?
        """

        # Create the params to create the rule
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2022, 10, 29),
            'until': datetime.datetime(2022, 11, 1),
            'byhour': 10,
        }

        timezone = pytz.timezone('Europe/Kiev')
        format = '%Y-%m-%d %H:%M'

        # Create the rule
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler',
            time_zone=timezone,
        )

        # Assert the initial values
        # Europe/Kiev is UTC + 3 prior to 10/30/22. (10 Europe/Kiev meeting is 7 UTC.)
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2022, 10, 29, 7))
        self.assertEqual(
            fleming.convert_to_tz(rule.next_occurrence, timezone).strftime(format),
            '2022-10-29 10:00'
        )

        # Notice next occurrence jumps to UTC + 2 to reflect change from DST on early hours of 10/30.
        # Notice the converted date is still the expected 10am.
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2022, 10, 29, 7))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2022, 10, 30, 8))
        self.assertEqual(
            fleming.convert_to_tz(rule.next_occurrence, timezone).strftime(format),
            '2022-10-30 10:00'
        )

        # Notice UTC + 2 is here to stay.
        # Notice the converted date is still the expected 10am.
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2022, 10, 30, 8))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2022, 10, 31, 8))
        self.assertEqual(
            fleming.convert_to_tz(rule.next_occurrence, timezone).strftime(format),
            '2022-10-31 10:00'
        )

    def test_model_default_time_zone(self):
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'UTC')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2))

    def test_model_different_time_zone_daily(self):
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 4, 3))

    def test_model_different_time_zone_ahead_daily(self):
        """
        Checks a time zone that is ahead of utc
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2018, 6, 19),
            'byhour': 1,
            'until': datetime.datetime(2018, 6, 20),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('Europe/London'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'Europe/London')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2018, 6, 19, 0))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2018, 6, 19, 0))
        self.assertEqual(rule.next_occurrence, None)

        # Check using a past time
        with freeze_time('1-3-2017'):
            # Use the force flag to get the next date
            self.assertEqual(rule.get_next_occurrence(), None)
            self.assertEqual(rule.get_next_occurrence(force=True), datetime.datetime(2018, 7, 19, 0))

        # Check using a future time
        with freeze_time('1-3-2117'):
            # Use the force flag to get the next date
            self.assertEqual(rule.get_next_occurrence(), None)
            self.assertEqual(rule.get_next_occurrence(force=True), datetime.datetime(2018, 7, 19, 0))

    def test_model_different_time_zone_ahead_crossing_day_daily(self):
        """
        Checks a time zone that is ahead of utc
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2018, 6, 19),
            'byhour': 0,
            'until': datetime.datetime(2018, 6, 20),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('Europe/London'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'Europe/London')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2018, 6, 18, 23))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2018, 6, 18, 23))
        self.assertEqual(rule.next_occurrence, None)

    def test_model_different_time_zone_monthly(self):
        """
        Test a monthly first day of month rule to catch case of converting tz back
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': 1,
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 2, 3))

        # Cross dst
        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 4, 2, 2))

    def test_model_different_time_zone_end_of_month(self):
        """
        Test a monthly first day of month rule to catch case of converting tz back
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': -1,
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 2, 1, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 3, 1, 3))

        # Test crossing dst
        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 3, 1, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 4, 1, 2))

    def test_generate_dates(self):
        """
        Assert generate_dates returns the same values as get_dates.
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': -1,
            'until': datetime.datetime(2017, 5, 1, 22),
        }

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(
            rule.get_dates(),
            rule.generate_dates()
        )

    def test_get_dates(self):
        """
        Test a monthly 1st day of month rule to catch case of converting tz back using the get_dates method
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2016, 12, 31),
            'bymonthday': 1,
            'until': datetime.datetime(2017, 4, 30),
        }

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )
        next_dates = rule.get_dates()

        # Check a few dates
        self.assertEqual(len(next_dates), 4)

        self.assertEqual(next_dates[0], datetime.datetime(2017, 1, 1, 5))
        self.assertEqual(next_dates[1], datetime.datetime(2017, 2, 1, 5))
        self.assertEqual(next_dates[2], datetime.datetime(2017, 3, 1, 5))
        self.assertEqual(next_dates[3], datetime.datetime(2017, 4, 1, 4))  # DST change for US/Eastern

        # Run pre save again to make sure it doesn't mess up params
        rule.pre_save_hooks()

        # Get next dates to compare against
        more_next_dates = rule.get_dates()
        self.assertEqual(next_dates, more_next_dates)

    def test_get_dates_with_start_date(self):
        """
        Test a date generation with a start date.
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2016, 12, 31),
            'bymonthday': 1,
        }

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )
        next_dates = rule.get_dates(num_dates=10, start_date=datetime.datetime(2018, 1, 1))

        # Check a few dates
        self.assertEqual(len(next_dates), 10)
        self.assertEqual(next_dates[0], datetime.datetime(2018, 1, 1, 5))
        self.assertEqual(next_dates[1], datetime.datetime(2018, 2, 1, 5))
        self.assertEqual(next_dates[2], datetime.datetime(2018, 3, 1, 5))
        self.assertEqual(next_dates[3], datetime.datetime(2018, 4, 1, 4))  # DST change for US/Eastern
        self.assertEqual(next_dates[-1], datetime.datetime(2018, 10, 1, 4))

    def test_get_dates_num_dates_greater(self):
        """
        Test a date generation with a start date and end date that will yield fewer dates than num_dates.
        Daily from 1/1 to 1/10 is 10. Request default of 20 but only after 1/5 which should yield 6 dates, inclusive.
        """
        rule = RRule(
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2017, 1, 1),
                'until': datetime.datetime(2017, 1, 10),
            },
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )
        next_dates = rule.get_dates(
            num_dates=20,
            start_date=datetime.datetime(2017, 1, 5)
        )

        self.assertEqual(len(next_dates), 6)
        self.assertEqual(next_dates[0], datetime.datetime(2017, 1, 5, 5))
        self.assertEqual(next_dates[1], datetime.datetime(2017, 1, 6, 5))
        self.assertEqual(next_dates[2], datetime.datetime(2017, 1, 7, 5))
        self.assertEqual(next_dates[3], datetime.datetime(2017, 1, 8, 5))
        self.assertEqual(next_dates[4], datetime.datetime(2017, 1, 9, 5))
        self.assertEqual(next_dates[5], datetime.datetime(2017, 1, 10, 5))

    def test_generate_dates_from_params(self):
        """
        Assert generate_dates_from_params returns the same values as get_dates_from_params.
        """
        params = {
            'rrule_params': {
                'freq': rrule.MONTHLY,
                'interval': 1,
                'dtstart': datetime.datetime(2017, 1, 1, 22),
                'bymonthday': -1,
                'until': datetime.datetime(2017, 5, 1, 22),
            },
            'time_zone': pytz.timezone('US/Eastern'),
            'num_dates': 3,
        }

        self.assertEqual(
            RRule.get_dates_from_params(**params),
            RRule.generate_dates_from_params(**params)
        )

    def test_get_dates_from_params(self):
        """
        Tests the class method wrapper
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': -1,
            'until': datetime.datetime(2017, 5, 1, 22),
        }

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )
        next_dates = rule.get_dates(
            num_dates=3,
            start_date=datetime.datetime(2017, 2, 2, 22)
        )

        next_dates_from_params = RRule.get_dates_from_params(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            num_dates=3,
            start_date=datetime.datetime(2017, 2, 2, 22)
        )

        self.assertEqual(next_dates, next_dates_from_params)

    def test_model_different_time_zone_end_of_month_get_dates(self):
        """
        Test a monthly first day of month rule to catch case of converting tz back using the get_dates method
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'bymonthday': -1,
        }

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )
        next_dates = rule.get_dates()
        self.assertEqual(next_dates[0], datetime.datetime(2017, 2, 1, 3))
        self.assertEqual(next_dates[1], datetime.datetime(2017, 3, 1, 3))
        self.assertEqual(next_dates[2], datetime.datetime(2017, 4, 1, 2))
        self.assertEqual(next_dates[3], datetime.datetime(2017, 5, 1, 2))
        self.assertEqual(next_dates[4], datetime.datetime(2017, 6, 1, 2))
        self.assertEqual(next_dates[5], datetime.datetime(2017, 7, 1, 2))
        self.assertEqual(next_dates[6], datetime.datetime(2017, 8, 1, 2))
        self.assertEqual(next_dates[7], datetime.datetime(2017, 9, 1, 2))
        self.assertEqual(next_dates[8], datetime.datetime(2017, 10, 1, 2))
        self.assertEqual(next_dates[9], datetime.datetime(2017, 11, 1, 2))
        self.assertEqual(next_dates[10], datetime.datetime(2017, 12, 1, 3))
        self.assertEqual(next_dates[11], datetime.datetime(2018, 1, 1, 3))
        self.assertEqual(next_dates[12], datetime.datetime(2018, 2, 1, 3))

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 2, 1, 3))

    def test_model_different_time_zone_daily_with_ending_early(self):
        """
        Makes sure the time_zone is respected and ends before the ending time
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'until': datetime.datetime(2017, 1, 3, 21),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3, 3))
        self.assertEqual(rule.next_occurrence, None)

        # Save again to hit all conditions (resaving when dtstart and until are already strings)
        rule.save()

        self.assertEqual(rule.rrule_params['dtstart'], '2017-01-01 22:00:00')
        self.assertEqual(rule.rrule_params['until'], '2017-01-03 21:00:00')

    def test_model_different_time_zone_daily_with_ending_on_interval(self):
        """
        Makes sure the time_zone is respected and ends before the ending time
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1, 22),
            'until': datetime.datetime(2017, 1, 3, 22),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.time_zone.zone, 'US/Eastern')
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3, 3))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 4, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 4, 3))
        self.assertEqual(rule.next_occurrence, None)

    def test_get_next_occurrence_count(self):
        """
        Makes sure the count ending is respected
        """
        params = {
            'freq': rrule.DAILY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 1),
            'count': 3,
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 2))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 3))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 3))
        self.assertEqual(rule.next_occurrence, None)

    @freeze_time('1-1-2016')
    def test_update_next_occurrence_ignore(self):
        """
        Make sure the next occurrence does not advance if the current time is beyond the next occurrence
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

    @freeze_time('1-1-2018')
    def test_update_next_occurrence(self):
        """
        Make sure the next occurrence does not advance if the current time is beyond the next occurrence
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        self.assertEqual(rule.last_occurrence, None)
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1))

        # Handle the next occurrence
        rule.update_next_occurrence()
        self.assertEqual(rule.last_occurrence, datetime.datetime(2017, 1, 1))
        self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 2))

    def test_refresh_next_occurrence(self):
        """
        Checks that the next occurrence is set relative to the passed occurrence (or last occurrence if null)
        Also verifies that a null next occurrence does not raise an exception, but sets the next to null
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2022, 2, 15),
            'until': datetime.datetime(2022, 2, 17),
        }

        with freeze_time('2-14-2022'):
            rule = RRule.objects.create(
                rrule_params=params,
                occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
            )
            self.assertEqual(rule.last_occurrence, None)
            self.assertEqual(rule.next_occurrence, datetime.datetime(2022, 2, 15))

        with freeze_time('2-15-2022'):
            rule.refresh_next_occurrence()
            self.assertEqual(rule.last_occurrence, None)
            self.assertEqual(rule.next_occurrence, datetime.datetime(2022, 2, 16))

        with freeze_time('2-16-2022'):
            rule.refresh_next_occurrence()
            self.assertEqual(rule.last_occurrence, None)
            self.assertEqual(rule.next_occurrence, datetime.datetime(2022, 2, 17))

        with freeze_time('2-17-2022'):
            rule.refresh_next_occurrence()
            self.assertEqual(rule.last_occurrence, None)
            self.assertEqual(rule.next_occurrence, None)

    def test_save_existing_update_next_occurrence(self):
        """
        When saving an existing rrule model, make sure that the next occurrence is updated when the
        next occurrence has not yet occurred, and the next occurrence is not updated if the start time
        is before the current date
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2017, 1, 1),
        }

        with freeze_time('1-1-2018'):
            rule = RRule.objects.create(
                rrule_params=params,
                occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler',
                time_zone=pytz.timezone('US/Eastern'),
            )

            self.assertEqual(rule.last_occurrence, None)
            self.assertEqual(rule.next_occurrence, datetime.datetime(2017, 1, 1, 5))

        # Save the rrule with a current time before the first occurrence
        with freeze_time('1-1-2016'):
            # Change the start date to a future date
            rule.rrule_params['dtstart'] = datetime.datetime(2018, 1, 1)
            rule.refresh_next_occurrence()
            rule.save()
            self.assertEqual(rule.next_occurrence, datetime.datetime(2018, 1, 1, 5))

            # Change the start date to a previous date that is still after the current date
            rule.rrule_params['dtstart'] = datetime.datetime(2016, 2, 1)
            rule.refresh_next_occurrence()
            rule.save()
            self.assertEqual(rule.next_occurrence, datetime.datetime(2016, 2, 1, 5))

            # Try setting the start time to a previous date before the current date and make sure it
            # gets set to the first occurrence after today
            rule.rrule_params['dtstart'] = datetime.datetime(2015, 12, 1)
            rule.refresh_next_occurrence()
            rule.save()
            self.assertEqual(rule.next_occurrence, datetime.datetime(2016, 1, 1, 5))

            # Coverage to test the string dtstart
            rule.rrule_params['dtstart'] = datetime.datetime(2015, 12, 1)
            rule.rrule_params['dtstart'] = rule.rrule_params['dtstart'].strftime('%Y-%m-%d %H:%M:%S')
            rule.save()
            self.assertEqual(rule.next_occurrence, datetime.datetime(2016, 1, 1, 5))

            # Refresh with date in past and make sure it is ignored
            rule.refresh_next_occurrence(current_time=datetime.datetime(2014, 1, 1))
            rule.save()
            self.assertEqual(rule.next_occurrence, datetime.datetime(2016, 1, 1, 5))

    def test_save_with_datetimes(self):
        """
        Verifies that the save method can work with dtstart and until params as datetime objects
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': datetime.datetime(2019, 5, 1),
            'until': datetime.datetime(2019, 6, 1),
        }
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler',
            time_zone=pytz.timezone('US/Eastern'),
        )

        self.assertEqual(rule.next_occurrence, datetime.datetime(2019, 5, 1, 4))
        self.assertEqual(rule.rrule_params['dtstart'], '2019-05-01 00:00:00')
        self.assertEqual(rule.rrule_params['until'], '2019-06-01 00:00:00')

    def test_save_with_strings(self):
        """
        Verifies that the save method can work with dtstart and until params as strings
        """
        params = {
            'freq': rrule.DAILY,
            'dtstart': '2019-05-01 00:00:00',
            'until': '2019-06-01 00:00:00',
        }
        rule = RRule.objects.create(
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler',
            time_zone=pytz.timezone('US/Eastern'),
        )

        self.assertEqual(rule.next_occurrence, datetime.datetime(2019, 5, 1, 4))
        self.assertEqual(rule.rrule_params['dtstart'], '2019-05-01 00:00:00')
        self.assertEqual(rule.rrule_params['until'], '2019-06-01 00:00:00')

    @freeze_time('6-15-2022')
    def test_clone(self):
        # New object that starts next Wednesday
        # Weekly on MWF
        rule = RRule.objects.create(
            rrule_params={
                'freq': rrule.WEEKLY,
                'dtstart': datetime.datetime(2022, 6, 22),
                'byweekday': [0, 2, 4],
            },
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Create a clone of the object.
        clone = RRule.clone(rule)

        # Assert that the clone's next occurrence is the same.
        format = '%Y-%m-%d'
        self.assertEqual(rule.next_occurrence.strftime(format), clone.next_occurrence.strftime(format))

        # Assert that the clone's params are the same
        self.assertEqual(rule.rrule_params, clone.rrule_params)

        # Assert the generated dates are equal.
        self.assertEqual(rule.get_dates(num_dates=4), clone.get_dates(num_dates=4))

    @freeze_time('6-15-2022')
    def test_weekly_clone_with_offset(self):
        # New object that starts next Wednesday
        # Weekly on MWF
        rule = RRule.objects.create(
            rrule_params={
                'freq': rrule.WEEKLY,
                'dtstart': datetime.datetime(2022, 6, 21),  # Tuesday
                'byweekday': [0, 2, 4],
            },
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler'
        )

        # Create a clones of the object with a start date of 2 days into the future and 2 days into the past.
        future_clone = rule.clone_with_day_offset(2)
        past_clone = rule.clone_with_day_offset(-2)

        # Assert that the rule created here is unchanged but the clones reflect their offsets.
        format = '%Y-%m-%d'
        self.assertEqual(rule.next_occurrence.strftime(format), '2022-06-22')
        self.assertEqual(future_clone.next_occurrence.strftime(format), '2022-06-24')
        self.assertEqual(past_clone.next_occurrence.strftime(format), '2022-06-20')

        # Assert that the rule and clones all reflect MWF
        self.assertEqual(rule.rrule_params['byweekday'], [0, 2, 4])
        self.assertEqual(future_clone.rrule_params['byweekday'], [0, 2, 4])
        self.assertEqual(past_clone.rrule_params['byweekday'], [0, 2, 4])

        # Assert the generated dates are as expected.
        self.assertEqual(
            rule.get_dates(num_dates=4),
            [
                datetime.datetime(2022, 6, 22),  # Wednesday
                datetime.datetime(2022, 6, 24),  # Friday
                datetime.datetime(2022, 6, 27),  # Monday
                datetime.datetime(2022, 6, 29),  # Wednesday
            ]
        )

        # Two days after each date in the regular series.
        self.assertEqual(
            future_clone.get_dates(num_dates=4),
            [
                datetime.datetime(2022, 6, 24),  # Friday
                datetime.datetime(2022, 6, 26),  # Sunday
                datetime.datetime(2022, 6, 29),  # Wednesday
                datetime.datetime(2022, 7, 1),   # Friday
            ]
        )

        # Two days before each date in the regular series.
        self.assertEqual(
            past_clone.get_dates(num_dates=4),
            [
                datetime.datetime(2022, 6, 20),  # Monday
                datetime.datetime(2022, 6, 22),  # Wednesday
                datetime.datetime(2022, 6, 25),  # Saturday
                datetime.datetime(2022, 6, 27),  # Monday
            ]
        )

    @freeze_time('6-1-2022')
    def test_monthly_clone_with_offset(self):
        """
        Assert than an object with bynweekday data can be cloned with an offset.
        """

        # Second Monday from end of month every other month
        # Starts today, 6/1. First occurrence is 6/20. (Last Monday is 27th.)
        data = {
            'freq': rrule.MONTHLY,
            'interval': 2,
            'dtstart': '6/1/2022',
            'byhour': '0',
            'time_zone': 'UTC',
            'ends': RecurrenceEnds.NEVER,
            'repeat_by': 'DAY_OF_THE_WEEK_END',
            'bynweekday': '[[0, -2]]'
        }

        # Form is used to flex the bynweeday to byweekday + bysetpos conversion that occurs in its save().
        form = RecurrenceForm(data=data)
        self.assertTrue(form.is_valid())
        rule = form.save()

        # Create a clones of the object with a start date of 2 days into the future and 2 days into the past.
        future_clone = rule.clone_with_day_offset(2)
        past_clone = rule.clone_with_day_offset(-2)

        # Assert that the rule created here is unchanged but the clones reflect their offsets.
        format = '%Y-%m-%d'
        self.assertEqual(rule.next_occurrence.strftime(format), '2022-06-20')
        self.assertEqual(future_clone.next_occurrence.strftime(format), '2022-06-22')
        self.assertEqual(past_clone.next_occurrence.strftime(format), '2022-06-18')

    @freeze_time('6-15-2022')
    def test_clone_with_day_offset_with_finish_date(self):
        """
        Assert that the 'until' param is offset correctly
        """

        rule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'dtstart': datetime.datetime(2022, 10, 15),
                'until': datetime.datetime(2022, 10, 17)
            }
        )

        future_clone = rule.clone_with_day_offset(1)
        past_clone = rule.clone_with_day_offset(-1)

        # Assert the cloned until values are the same
        self.assertEqual(
            parser.parse(rule.rrule_params['until']),
            parser.parse(future_clone.rrule_params['until'])
        )
        self.assertEqual(
            parser.parse(rule.rrule_params['until']),
            parser.parse(past_clone.rrule_params['until'])
        )

        # Assert the generated dates are as expected for the original.
        self.assertEqual(
            rule.get_dates(),
            [
                datetime.datetime(2022, 10, 15),
                datetime.datetime(2022, 10, 16),
                datetime.datetime(2022, 10, 17)
            ]
        )

        # One day after each date in the regular series.
        self.assertEqual(
            future_clone.get_dates(),
            [
                datetime.datetime(2022, 10, 16),
                datetime.datetime(2022, 10, 17),
                datetime.datetime(2022, 10, 18)
            ]
        )

        # One day before each date in the regular series.
        self.assertEqual(
            past_clone.get_dates(),
            [
                datetime.datetime(2022, 10, 14),
                datetime.datetime(2022, 10, 15),
                datetime.datetime(2022, 10, 16)
            ]
        )

    @freeze_time('10-31-2022')
    def test_clone_with_future_days_across_dst(self):
        """
        Assert that the resulting recurrence next occurrence reflects *its* timezone offset.
        Fall Back.
        """

        rule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'dtstart': datetime.datetime(2022, 10, 29, 10),
                'until': datetime.datetime(2022, 11, 1, 10),
            },
            time_zone=pytz.timezone('Europe/Kiev')
        )

        # Europe/Kiev goes from UTC+3 to UTC+2 in the early hours of 10/30.
        self.assertEqual(
            rule.get_dates(),
            [
                datetime.datetime(2022, 10, 29, 7),
                datetime.datetime(2022, 10, 30, 8),
                datetime.datetime(2022, 10, 31, 8),
                datetime.datetime(2022, 11, 1, 8),
            ]
        )

        # Europe/Kiev goes to standard time (UTC+2) in the early hours of 1/30.
        # This offset should result in a datetime (UTC+3).
        past_clone = rule.clone_with_day_offset(-1)

        # One day before each date in the regular series.
        self.assertEqual(
            past_clone.get_dates(),
            [
                datetime.datetime(2022, 10, 28, 7),
                datetime.datetime(2022, 10, 29, 7),
                datetime.datetime(2022, 10, 30, 8),
                datetime.datetime(2022, 10, 31, 8),
            ]
        )

    @freeze_time('10-31-2022')
    def test_clone_with_offset_from_dst(self):
        """
        Assert that a clone, with a negative offset results in dates that respect a DST transition.
        Europe/Kiev goes from UTC+3 to UTC+2 in the early hours of 10/30.
        Fall back.
        """

        # Starting in standard time
        rule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'dtstart': datetime.datetime(2022, 10, 31, 10),
                'until': datetime.datetime(2022, 11, 3, 10),
            },
            time_zone=pytz.timezone('Europe/Kiev'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne',
        )

        # Times are UTC+2 because it is after the 10/30 transition.
        self.assertEqual(
            rule.generate_dates(),
            [
                datetime.datetime(2022, 10, 31, 8),
                datetime.datetime(2022, 11, 1, 8),
                datetime.datetime(2022, 11, 2, 8),
                datetime.datetime(2022, 11, 3, 8),
            ]
        )

        # Clone prior to the DST switch and ensure the times transition between 10/30 & 10/31.
        past_clone = rule.clone_with_day_offset(-3)
        expected_clone_dates = [
            datetime.datetime(2022, 10, 28, 7),
            datetime.datetime(2022, 10, 29, 7),
            datetime.datetime(2022, 10, 30, 8),
            datetime.datetime(2022, 10, 31, 8),
        ]
        self.assertEqual(past_clone.generate_dates(), expected_clone_dates)

        # Progress through dates and assert generated next_occurrence matches get_dates()
        for x, expected_date in enumerate(expected_clone_dates):
            with freeze_time(expected_date):
                RRule.objects.handle_overdue()
                past_clone.refresh_from_db()
                if x < len(expected_clone_dates) - 1:
                    self.assertEqual(
                        past_clone.next_occurrence,
                        expected_clone_dates[x + 1],
                    )

    @freeze_time('10-31-2022')
    def test_clone_with_offset_to_dst(self):
        """
        Assert that a clone, with a negative offset results in dates that respect a DST transition.
        Europe/Kiev goes from UTC+2 to UTC+3 in the early hours of 3/27.
        Spring forward.
        """

        # Starting in standard time
        rule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'dtstart': datetime.datetime(2022, 3, 28, 10),
                'until': datetime.datetime(2022, 3, 31, 10),
            },
            time_zone=pytz.timezone('Europe/Kiev'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne',
        )

        # Times are UTC+2 because it is after the 3/27 transition.
        self.assertEqual(
            rule.generate_dates(),
            [
                datetime.datetime(2022, 3, 28, 7),
                datetime.datetime(2022, 3, 29, 7),
                datetime.datetime(2022, 3, 30, 7),
                datetime.datetime(2022, 3, 31, 7),
            ]
        )

        # Clone prior to the DST switch and ensure the times transition between 3/26 & 3/27.
        past_clone = rule.clone_with_day_offset(-3)
        expected_clone_dates = [
            datetime.datetime(2022, 3, 25, 8),
            datetime.datetime(2022, 3, 26, 8),
            datetime.datetime(2022, 3, 27, 7),
            datetime.datetime(2022, 3, 28, 7),
        ]
        self.assertEqual(past_clone.generate_dates(), expected_clone_dates)

        # Progress through dates and assert generated next_occurrence matches get_dates()
        for x, expected_date in enumerate(expected_clone_dates):
            with freeze_time(expected_date):
                RRule.objects.handle_overdue()
                past_clone.refresh_from_db()
                if x < len(expected_clone_dates) - 1:
                    self.assertEqual(
                        past_clone.next_occurrence,
                        expected_clone_dates[x + 1],
                    )

    def test_offset(self):
        """
        Assert the offset method adjusts a given date by the given number of days
        """
        self.assertEqual(
            RRule(day_offset=1).offset(datetime.datetime(2023, 1, 1)),
            datetime.datetime(2023, 1, 2)
        )

        self.assertEqual(
            RRule(day_offset=-1).offset(datetime.datetime(2023, 1, 1)),
            datetime.datetime(2022, 12, 31)
        )

        self.assertEqual(
            RRule().offset(datetime.datetime(2023, 1, 1)),
            datetime.datetime(2023, 1, 1)
        )

    def test_day_offset(self):
        """
        Assert that the field, day_offset, adjusts all generated dates.
        """
        params = {
            'freq': rrule.MONTHLY,
            'interval': 1,
            'dtstart': datetime.datetime(2016, 12, 31),
            'bymonthday': 1,
            'until': datetime.datetime(2017, 4, 30),
        }

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler',
            day_offset=1,
        )
        next_dates = rule.get_dates()

        # Check a few dates
        self.assertEqual(len(next_dates), 4)

        self.assertEqual(next_dates[0], datetime.datetime(2017, 1, 2, 5))
        self.assertEqual(next_dates[1], datetime.datetime(2017, 2, 2, 5))
        self.assertEqual(next_dates[2], datetime.datetime(2017, 3, 2, 5))
        self.assertEqual(next_dates[3], datetime.datetime(2017, 4, 2, 4))  # DST change for US/Eastern

        # Run pre save again to make sure it doesn't mess up params
        rule.pre_save_hooks()

        # Get next dates to compare against
        more_next_dates = rule.get_dates()
        self.assertEqual(next_dates, more_next_dates)

        rule = RRule(
            rrule_params=params,
            time_zone=pytz.timezone('US/Eastern'),
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.MockHandler',
            day_offset=-1,
        )

        next_dates = rule.get_dates()

        # Check a few dates
        self.assertEqual(len(next_dates), 4)

        self.assertEqual(next_dates[0], datetime.datetime(2016, 12, 31, 5))
        self.assertEqual(next_dates[1], datetime.datetime(2017, 1, 31, 5))
        self.assertEqual(next_dates[2], datetime.datetime(2017, 2, 28, 5))
        self.assertEqual(next_dates[3], datetime.datetime(2017, 3, 31, 4))  # DST change for US/Eastern

        # Run pre save again to make sure it doesn't mess up params
        rule.pre_save_hooks()

        # Get next dates to compare against
        more_next_dates = rule.get_dates()
        self.assertEqual(next_dates, more_next_dates)

    def test_next_occurrences_day_offset(self):
        """
        Make sure that the correct occurrences are selected when a day offset is configured.
        """
        params = {
            'freq': rrule.WEEKLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        rrule1 = G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne',
            day_offset=1
        )

        params = {
            'freq': rrule.WEEKLY,
            'interval': 1,
            'dtstart': datetime.datetime(2017, 1, 2),
        }

        rrule2 = G(
            RRule,
            rrule_params=params,
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerTwo',
            day_offset=-1
        )

        self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 3))
        self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 1))

        with freeze_time('1-3-2017'):
            RRule.objects.handle_overdue()

            rrule1.refresh_from_db()
            rrule2.refresh_from_db()

            # Both should be progressed
            self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 10))
            self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 8))

        with freeze_time('1-8-2017'):
            RRule.objects.handle_overdue()

            rrule1.refresh_from_db()
            rrule2.refresh_from_db()

            # Both should be progressed
            self.assertEqual(rrule1.next_occurrence, datetime.datetime(2017, 1, 10))
            self.assertEqual(rrule2.next_occurrence, datetime.datetime(2017, 1, 15))

    def test_next_occurrence_clone_with_offset(self):
        """
        Cloning an established object with an offset should yield a next_occurrence of the object's with the offset.
        """
        # A daily recurrence with a multi-day offset to highlight differences clearly.
        rrule1 = G(
            RRule,
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2023, 1, 1),
            },
            occurrence_handler_path='ambition_utils.rrule.tests.model_tests.HandlerOne',
        )

        # Many days into the future
        with freeze_time('1-10-2023'):
            # Catch the recurrence object up.
            while rrule1.next_occurrence <= datetime.datetime.now():
                RRule.objects.handle_overdue()
                rrule1.refresh_from_db()
            self.assertEqual(rrule1.next_occurrence, datetime.datetime(2023, 1, 11))

            # Create a clone with an offset and assert that the clone's next occurrence is 1/11 -5 = 1/6.
            rrule2 = rrule1.clone_with_day_offset(day_offset=-5)
            self.assertEqual(rrule2.next_occurrence, datetime.datetime(2023, 1, 6))

            # First object is unchanged
            rrule1.refresh_from_db()
            self.assertEqual(rrule1.next_occurrence, datetime.datetime(2023, 1, 11))


class RRuleWithExclusionTest(TestCase):
    def test_exclusion(self):
        weekly_program = Program.objects.create(name='Weekly Program')
        weekly_rrule = RRule.objects.create(
            rrule_params={
                'freq': rrule.DAILY,
                'interval': 1,
                'dtstart': datetime.datetime(2023, 6, 1),
                'byweekday': 0,
            },
            rrule_exclusion_params={
                'freq': rrule.MONTHLY,
                'interval': 1,
                'dtstart': datetime.datetime(2023, 6, 1),
                'bysetpos': 2,
                'byweekday': 0
            },
            related_object=weekly_program,
            related_object_handler_name='handle_start_recurrence',
        )
        monthly_program = Program.objects.create(name='Monthly Program')
        monthly_rrule = RRule.objects.create(
            rrule_params={
                'freq': rrule.MONTHLY,
                'interval': 1,
                'dtstart': datetime.datetime(2023, 6, 1),
                'bysetpos': 2,
                'byweekday': 0
            },
            related_object=monthly_program,
            related_object_handler_name='handle_start_recurrence',
        )

        weekly_dates = weekly_rrule.get_dates(num_dates=10)
        monthly_dates = monthly_rrule.get_dates(num_dates=10)
        self.assertEqual(
            weekly_dates,
            [datetime.datetime(2023, 6, 5, 0, 0), datetime.datetime(2023, 6, 19, 0, 0),
             datetime.datetime(2023, 6, 26, 0, 0), datetime.datetime(2023, 7, 3, 0, 0),
             datetime.datetime(2023, 7, 17, 0, 0), datetime.datetime(2023, 7, 24, 0, 0),
             datetime.datetime(2023, 7, 31, 0, 0), datetime.datetime(2023, 8, 7, 0, 0),
             datetime.datetime(2023, 8, 21, 0, 0), datetime.datetime(2023, 8, 28, 0, 0)]
        )
        self.assertEqual(
            monthly_dates,
            [datetime.datetime(2023, 6, 12, 0, 0), datetime.datetime(2023, 7, 10, 0, 0),
             datetime.datetime(2023, 8, 14, 0, 0), datetime.datetime(2023, 9, 11, 0, 0),
             datetime.datetime(2023, 10, 9, 0, 0), datetime.datetime(2023, 11, 13, 0, 0),
             datetime.datetime(2023, 12, 11, 0, 0), datetime.datetime(2024, 1, 8, 0, 0),
             datetime.datetime(2024, 2, 12, 0, 0), datetime.datetime(2024, 3, 11, 0, 0)]
        )



# Generated by Django 2.2.28 on 2023-01-06 21:15

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='FloatModel',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('value', models.FloatField()),
            ],
        ),
        migrations.CreateModel(
            name='IncrementalAnomaly',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(max_length=256, unique=True)),
                ('blob', models.JSONField(default=dict, null=True)),
                ('percentile_low', models.FloatField(null=True)),
                ('percentile_high', models.FloatField(null=True)),
                ('delta', models.FloatField(default=0.01, null=True)),
                ('K', models.FloatField(default=25, null=True)),
                ('threshold_low', models.FloatField(db_index=True, null=True)),
                ('threshold_high', models.FloatField(db_index=True, null=True)),
                ('num_values_ingested', models.IntegerField(default=0)),
                ('last_modified', models.DateTimeField(db_index=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NonIncrementalAnomaly',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uid', models.CharField(max_length=256, unique=True)),
                ('blob', models.JSONField(default=dict, null=True)),
                ('percentile_low', models.FloatField(null=True)),
                ('percentile_high', models.FloatField(null=True)),
                ('delta', models.FloatField(default=0.01, null=True)),
                ('K', models.FloatField(default=25, null=True)),
                ('threshold_low', models.FloatField(db_index=True, null=True)),
                ('threshold_high', models.FloatField(db_index=True, null=True)),
                ('num_values_ingested', models.IntegerField(default=0)),
                ('last_modified', models.DateTimeField(db_index=True)),
                ('name', models.CharField(max_length=100)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]



import pytz
from django.test import TestCase, SimpleTestCase
from freezegun import freeze_time

from ambition_utils.time_helpers import get_time_zones, Weekday


class TimeHelperTestCase(TestCase):

    @freeze_time('1-1-2017')
    def test_get_time_zones(self):
        """
        Makes sure the time zones are returned as dicts and that eastern is first
        """

        time_zones = get_time_zones()

        self.assertEqual(time_zones[0]['id'], 'US/Eastern')
        self.assertEqual(time_zones[0]['name'], 'US/Eastern (EST) (GMT -5)')
        self.assertTrue(len(time_zones) > 400)

    @freeze_time('1-1-2017')
    def test_get_time_zones_as_tuple(self):
        """
        Makes sure the time zones are returned as tuples and that eastern is first
        """

        time_zones = get_time_zones(return_as_tuple=True)

        self.assertEqual(time_zones[0][0], 'US/Eastern')
        self.assertEqual(time_zones[0][1], 'US/Eastern (EST) (GMT -5)')

    def test_get_timezones_contains_all_timezones(self):
        """
        Ensures that the get_timezones method returns pytz.all_timezones rather than pytz.common_timezones
        """
        # get a set of all_timezones from pytz
        timezones_all_pytz = set(pytz.all_timezones)

        # get the actual timezones we are returning
        timezones_actual = get_time_zones(return_as_tuple=False)
        timezones_actual_set = set()

        # grab the appropriate timezone format for comparison
        [timezones_actual_set.add(item['id']) for item in timezones_actual]

        # grab the timezones that exist in pytz.all_timezones but don't exist in the actual timezones set
        timezones_difference = timezones_all_pytz.difference(timezones_actual_set)

        self.assertEqual(len(timezones_difference), 0)


class TestWeekday(SimpleTestCase):
    def test_bad_convention(self):
        with self.assertRaises(ValueError):
            Weekday(0, convention='bad')

    def test_bad_day(self):
        with self.assertRaises(ValueError):
            Weekday(9)

    def test_mondays(self):

        mondays = {
            'python': 0,
            'django': 2,
            'postgres': 1,
            'iso': 1,
        }

        sundays = {
            'python': 6,
            'django': 1,
            'postgres': 0,
            'iso': 7,
        }

        # Test that mondays look right
        python_day = 0
        weekday = Weekday(python_day)
        for convention in mondays.keys():
            # Test forward
            convention_day1 = weekday[convention]
            convention_day2 = getattr(weekday, convention)
            self.assertEqual(convention_day1, convention_day2)
            self.assertEqual(convention_day1, mondays[convention])

            # Test backward
            weekday_back = Weekday(convention_day1, convention)
            self.assertEqual(weekday_back.python, python_day)

        # Test that sundays look right
        python_day = 6
        weekday = Weekday(python_day)
        for convention in mondays.keys():
            convention_day1 = weekday[convention]
            convention_day2 = getattr(weekday, convention)
            self.assertEqual(convention_day1, convention_day2)
            self.assertEqual(convention_day1, sundays[convention])

            # Test backward
            weekday_back = Weekday(convention_day1, convention)
            self.assertEqual(weekday_back.python, python_day)



from django.test import TestCase

from ambition_utils.activity.models import ActivityGroup


class ActivityGroupTest(TestCase):

    def test_str(self):
        self.assertEqual(str(ActivityGroup(name='test name')), 'test name')



from datetime import datetime
from pytz import all_timezones, timezone, exceptions as pytz_exceptions


def get_gmt_offset(tz_name, now):
    try:
        offset = timezone(tz_name).utcoffset(now)
    except pytz_exceptions.AmbiguousTimeError:  # pragma: no cover
        offset = timezone(tz_name).utcoffset(now, is_dst=False)
    except pytz_exceptions.NonExistentTimeError:  # pragma: no cover
        offset = timezone(tz_name).utcoffset(now, is_dst=False)

    offset = int(offset.total_seconds() / 3600)

    return '{0}'.format(offset) if offset < 0 else '+{0}'.format(offset)


def get_time_zones(return_as_tuple=False):
    """
    Gets timezones to display to the front end (and for validation). Orders timezones
    with common US ones first and attaches GMT offset to them.
    """
    us_tzs = [
        ('US/Eastern', 'US/Eastern (EST)'), ('US/Central', 'US/Central (CST)'),
        ('US/Mountain', 'US/Mountain (MST)'), ('US/Pacific', 'US/Pacific (PST)'),
        ('US/Arizona', 'US/Arizona'), ('US/Alaska', 'US/Alaska'), ('US/Hawaii', 'US/Hawaii')
    ]
    other_tzs = set(all_timezones) - set(us_tz[0] for us_tz in us_tzs)
    all_tzs = us_tzs + [(tz, tz) for tz in other_tzs]

    # Attach GMT values to the display names of the tzs
    now = datetime.utcnow()
    all_tzs = [
        (tz[0], '{0} (GMT {1})'.format(tz[1], get_gmt_offset(tz[0], now))) for tz in all_tzs
    ]

    return all_tzs if return_as_tuple else [{
        'id': tz[0],
        'name': tz[1],
    } for tz in all_tzs]


class Weekday:
    """
    Python, Postgres and Django each have different conventions for assigning numbers to weekdays.
    This class provides a general utility for translating between the conventions when you need to.

    See the following for references to the different conventions:
        - Python datetime.weekday() and datetime.isoweekday(): https://docs.python.org/3.6/library/datetime.html
        - Django: https://docs.djangoproject.com/en/dev/ref/models/querysets/#week-day
        - Postgres DOW() function: https://www.postgresql.org/docs/8.2/static/functions-datetime.html

    The translations are implemented using a lookup table.  There is probably some fancy modulo math
    that could be used, but lookups are fairly quick, and easy to understand.

    Here are some examples:


    # Convert Python Monday to a Postgres Monday
    postgres_monday = Weekday(0, convention='python').postgres
    postgres_monday = Weekday(0, convention='python')['postgres']

    # Convert a Postgres Monday to a Python Monday
    python_monday = Weekday(1, convention='postgres').python

    """
    # This lookup has keys of the python convention and values of each of the other types of conventions
    LOOKUP = {
        ('python', 'python'): {0: 0, 1: 1, 2: 2, 3: 3, 4: 4, 5: 5, 6: 6},
        ('python', 'django'): {0: 2, 1: 3, 2: 4, 3: 5, 4: 6, 5: 7, 6: 1},
        ('python', 'postgres'): {0: 1, 1: 2, 2: 3, 3: 4, 4: 5, 5: 6, 6: 0},
        ('python', 'iso'): {0: 1, 1: 2, 2: 3, 3: 4, 4: 5, 5: 6, 6: 7}
    }

    # Compute reverse relations
    LOOKUP[('django', 'python')] = {v: k for (k, v) in LOOKUP[('python', 'django')].items()}
    LOOKUP[('postgres', 'python')] = {v: k for (k, v) in LOOKUP[('python', 'postgres')].items()}
    LOOKUP[('iso', 'python')] = {v: k for (k, v) in LOOKUP[('python', 'iso')].items()}

    _CONVENTIONS = {t[1] for t in LOOKUP.keys()}

    @classmethod
    def _check_convention(cls, convention):
        """
        Make sure only valid conventions are specified
        """
        if convention not in cls._CONVENTIONS:
            raise ValueError(f'Allowed conventions: {cls._CONVENTIONS}')

    def __init__(self, input_day, convention='python'):
        """
        :param input_day: An integer representing the day
        :param convention:  The convention assumed for the input day
        """
        # Make sure the convention is valid
        self._check_convention(convention)

        # Get the lookup from the input convention into Python
        to_python = self.LOOKUP[(convention, 'python')]

        # Make sure the input day is valid
        if input_day not in to_python:
            raise ValueError(f'Valid input days for {convention} are {list(to_python.keys())}')

        # Convert the input into a python weekday
        python_day = self.LOOKUP[(convention, 'python')][input_day]

        # Populate the appropriate values for all the conventions
        for conv in self._CONVENTIONS:
            setattr(self, conv, self.LOOKUP[('python', conv)][python_day])

    def __getitem__(self, convention):
        """
        Allow for dictionary-like access to convention attributes
        """
        self._check_convention(convention)
        return getattr(self, convention)



import logging

import sys
from django.db import connection, transaction
from django.db.utils import OperationalError

from ambition_utils.postgres_lock.models import PostgresLock


LOG = logging.getLogger(__name__)


class PostgresLockException(Exception):
    """
    An exception that is raised if there is an error trying to acquire the lock
    """
    pass


class PostgresLockContext(object):
    """
    Context manager for a postgres lock
    """

    def __init__(self, name, value=None, timeout=60 * 15):
        # Save the name
        self._name = name

        # Save the value
        self._value = value

        # Create an empty transaction
        self._transaction = None

        # Set the timeout
        self._timeout = '{0}s'.format(timeout)

        # Set a place to store the lock
        self._lock = None

        # Call the parent
        super(PostgresLockContext, self).__init__()

    @property
    def lock(self):
        return self._lock

    @property
    def transaction(self):
        return self._transaction

    def __enter__(self):
        # Log that we are trying to acquire the lock
        LOG.info('Waiting to acquire lock: {0}'.format(self._name))

        # Create the transaction
        self._transaction = transaction.atomic()

        # Build the query
        query = """
        INSERT INTO {table}(
                name,
                time,
                value
            )
            VALUES (
                %(name)s,
                now(),
                %(value)s
            )
            ON CONFLICT (name) DO UPDATE
            SET
                time = now(),
                value = excluded.value,
                previous_value = {table}.value
            RETURNING
                name,
                time,
                value,
                previous_value;
        """.format(
            table=PostgresLock._meta.db_table,
        )

        # Start the transaction
        self._transaction.__enter__()

        # Keep a reference to the exception
        exception = None

        # Create the connection
        with connection.cursor() as cursor:
            # Get the default timeout
            cursor.execute('SHOW statement_timeout')
            default_timeout = cursor.fetchone()[0]

            # Set the timeout
            cursor.execute('SET statement_timeout = %(timeout)s', {
                'timeout': self._timeout
            })

            # Acquire the lock
            try:
                cursor.execute(
                    query,
                    {
                        'name': self._name,
                        'value': self._value
                    }
                )
                self._lock = dict(zip([col[0] for col in cursor.description], cursor.fetchone()))
            except OperationalError:
                exception = PostgresLockException('Timed out waiting for lock')

            # Reset the timeout
            if exception is None:
                cursor.execute('SET statement_timeout = %(timeout)s', {
                    'timeout': default_timeout
                })

        # If we have an exception, raise it
        if exception is not None:
            self.__exit__(*sys.exc_info())
            raise exception

        # At this point we have the lock
        LOG.info('Successfully acquired lock: {0}'.format(self._name))

        # Return self
        return self

    def __exit__(self, *args, **kwargs):
        # Complete the transaction
        if self._transaction:
            self._transaction.__exit__(*args, **kwargs)
        else:  # pragma: no cover
            pass

    def values_match(self):
        if self._lock is None:
            return False
        return self._lock['value'] == self._lock['previous_value']



import os
import inspect
from collections import namedtuple
from django.template import Template, Context
from django.db.utils import ProgrammingError
from typing import Dict, List, Tuple, Any


def queryset_to_sql(queryset):
    """
    Transform a queryset into pretty sql that can be copy-pasted directly
    into pg-admin
    """
    # Do imports here to avoid dependencies
    import sqlparse
    from django.db import connection

    # Compile the query to python db api
    sql, sql_params = queryset.query.get_compiler(using=queryset.db).as_sql()

    # Translate the python query spec into a postgres query
    with connection.cursor() as cur:
        query = cur.mogrify(sql, sql_params)

    # Make the query pretty and return it
    query = sqlparse.format(query, reindent=True, keyword_case='upper')
    return query


class SQLBase(object):

    def __init__(self):
        self._raw_sql = None
        self._params = []
        self._raw_results = None
        self._raw_columns = None
        self._raw_connection = None
        self._django_context = None
        self._raw_sql = None
        self._rendered_sql = None

    @property
    def raw_sql(self):
        if self._rendered_sql is None:
            if self._django_context is not None:
                template = Template(self._raw_sql)
                self._rendered_sql = template.render(context=Context(self._django_context))
            else:
                self._rendered_sql = self._raw_sql
        return self._rendered_sql

    @property
    def _connection(self):
        if self._raw_connection is None:
            from django.db import connection
            self._raw_connection = connection
        return self._raw_connection

    @property
    def _results(self):
        if self._raw_results is None:
            self.run()
        return self._raw_results

    @property
    def _columns(self):
        if self._raw_columns is None:
            self.run()
        return self._raw_columns

    def run(self):
        with self._connection.cursor() as cursor:
            cursor.execute(self.raw_sql, self._params)
            try:
                self._raw_results = list(cursor.fetchall())
                self._raw_columns = [col[0] for col in cursor.description]
            except ProgrammingError as e:
                if str(e) == 'no results to fetch':
                    self._raw_results = []
                    self._raw_columns = []
                else:  # pragma: no cover  No expected to hit this, but raise just in case
                    raise

    def using_connection(self, connection):
        self._raw_connection = connection
        return self

    def with_context(self, context: Dict[str, Any]) -> 'SQLBase':
        """
        specify a dict of django-context for rendering sql
        """
        self._django_context = context
        return self

    def with_params(self, params: Dict[str, Any]) -> 'SQLBase':
        """
        specify a list or dict of sql params
        """
        self._params = params
        return self

    def as_tuples(self) -> List[Tuple]:
        """
        :return: Results as a list of tuples
        """
        return self._results

    def as_dicts(self) -> List[Dict[str, Any]]:
        """
        :return: Results as a list of dicts
        """
        return [dict(zip(self._columns, row)) for row in self._results]

    def as_named_tuples(self, named_tuple_name='Result') -> List[Any]:
        """
        :return: Results as a list of named tuples
        """
        # Ignore typing in here because of unconventional namedtuple usage
        nt_result = namedtuple(named_tuple_name, self._columns)  # type: ignore
        return [nt_result(*row) for row in self._results]  # type: ignore

    def as_dataframe(self) -> Any:
        """
        :return: Results as a pandas dataframe
        """
        try:
            import pandas as pd
        except ImportError:  # pragma: no cover.  Not going to uninstall pandas to test this
            raise ImportError('\n\nNope! This method requires that pandas be installed.  You know what to do.')

        return pd.DataFrame(self._results, columns=self._columns)

    def to_tuples(self) -> List[Tuple]:
        """
        alias
        """
        return self.as_tuples()

    def to_dicts(self) -> List[Dict[str, Any]]:
        """
        alias
        """
        return self.as_dicts()

    def to_named_tuples(self) -> List[Any]:
        """
        alias
        """
        return self.as_named_tuples()

    def to_dataframe(self) -> Any:
        """
        alias
        """
        return self.as_dataframe()


class FileSQL(SQLBase):
    def __init__(
            self,
            path_to_sql_file: str,
            path_is_relative: bool = True
    ) -> None:
        """

        :param path_to_sql_file: the relative path from the file you call this in to the sql file you
                                 want to execute.
        :param params: The params that will get passed into the sql template you right.
                       See: https://www.python.org/dev/peps/pep-0249/#paramstyle

        Here is an example:
            Lets say you have directory structure
            my_app
              |
              +-- my_python_code
                     |
                     +-- my_script.py
              +-- my_sql_code
                     |
                     +-- query.sql

            Then in my_app/my_python_code/my_script.py

            # These can be optionally passed to the SimpleSql object
            connection = get_my_connection_function()
            params = get_my_params_list_or_dict()

            # run simple query with no params and return the results as dicts
            SQL('../my_sql_code.sql').as_dicts()

            # run sql with a custom connection and parameters retrieving results as dataframe
            # note that the abolute path to sql file is now specified
            sql = SQL('/Users/billybob/my_sql_code.sql', path_is_relative=False)
            sql.using(connection)
            sql.with_params(params)
            df = sql.as_dataframe()

            # exact same command as above, but chained as a one-liner
            results = SQL('../my_sql_code.sql').using(connection).with_params(params).as_dataframe()

            # print the raw sql template
            print(sql.raw_sql)
        """
        super(FileSQL, self).__init__()
        # get the path of the file that is calling this constructor
        frame = inspect.stack()[1]
        calling_module = inspect.getmodule(frame[0])
        calling_module_path = os.path.abspath(os.path.dirname(calling_module.__file__))

        # read the sql template from the file
        if path_is_relative:
            path_to_sql_file = os.path.join(calling_module_path, path_to_sql_file)

        self._path_to_sql_file = path_to_sql_file

        with open(os.path.realpath(path_to_sql_file)) as sql_file:
            self._raw_sql = sql_file.read()


class StringSQL(SQLBase):
    def __init__(
            self,
            sql_query_string: str
    ) -> None:
        """
        Simple wrapper to execute a sql query against django models
        and provide the results as different object types.
        """
        super(StringSQL, self).__init__()
        self._raw_sql = sql_query_string



# Generated by Django 3.2.13 on 2022-05-25 17:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('postgres_lock', '0002_auto_20191030_1417'),
    ]

    operations = [
        migrations.AddField(
            model_name='postgreslock',
            name='previous_value',
            field=models.TextField(default=None, null=True),
        ),
        migrations.AddField(
            model_name='postgreslock',
            name='value',
            field=models.TextField(default=None, null=True),
        ),
    ]



# Generated by Django 3.2.19 on 2023-06-05 19:55

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rrule', '0003_auto_20220623_1649'),
    ]

    operations = [
        migrations.AddField(
            model_name='rrule',
            name='rrule_exclusion_params',
            field=models.JSONField(blank=True, default=None, null=True),
        ),
    ]



from django.db import models


class Program(models.Model):
    name = models.TextField()
    start_called = models.IntegerField(default=0)
    end_called = models.IntegerField(default=0)

    start_recurrence = models.ForeignKey(
        'rrule.RRule', on_delete=models.SET_NULL, null=True,
        related_name='programs_start'
    )
    end_recurrence = models.ForeignKey(
        'rrule.RRule', on_delete=models.SET_NULL, null=True,
        related_name='programs_end'
    )

    def handle_start_recurrence(self, rrule):
        self.start_called += 1
        self.save()
        return rrule

    def handle_end_recurrence(self, rrule):
        self.end_called += 1
        self.save()
        return rrule



from django.db import transaction, utils
from django.test.testcases import TestCase
from unittest.mock import patch

from ambition_utils.transaction import durable
from ambition_utils.transaction import decorators
from ambition_utils.transaction.utils import DurableResetAtomicForTestingPatcher


class DurableTests(TestCase):
    def setUp(self):
        # Call the parent setup which starts the transaction
        super().setUp()

        # Patch our atomic so we think we have not yet started a transaction
        patcher = DurableResetAtomicForTestingPatcher()
        patcher.start()

        # Add cleanup to stop the patches
        self.addCleanup(patcher.stop)

    def test_nested_atomic_yields_exception(self):
        """
        Verifies that a function wrapped in transaction.durable will yield an exception in a nested transaction
        """

        @transaction.atomic
        def bad_consumer():
            print('Running func')
            test_function()

        self.assertRaises(utils.ProgrammingError, bad_consumer)

    def test_correct_usage(self):
        """
        Verifies that a function wrapped in transaction.durable will not be nested in a transaction
        """
        def test_durable_method():
            # Returns true if the function is in a transaction
            return decorators._is_in_atomic_block()

        self.assertFalse(test_function(test_durable_method))

    @patch('ambition_utils.transaction.decorators.transaction.rollback')
    @patch('ambition_utils.transaction.decorators._db_may_have_uncommitted_work')
    def test_hanging_transaction(self, mock_db_may_have_uncommitted_work, mock_rollback):
        """
        Verify that an exception is raised if we have uncommitted work
        """
        mock_db_may_have_uncommitted_work.return_value = True

        self.assertRaises(utils.ProgrammingError, test_function)
        mock_rollback.assert_called_once_with()


@durable
def test_function(callback=None):
    if callback is not None:
        return callback()



import pytz
from django.test import TestCase

from ambition_utils.fields import TimeZoneField
from ambition_utils.tests.models import FakeModel


class TimeZoneFieldTest(TestCase):
    """
    Verifies the time zone field is cast when being assigned
    """

    def test_cast_on_assign(self):
        """

        """
        instance = FakeModel()
        instance.cast_time_zone_field = 'US/Eastern'

        # Verify the timezone was cast
        self.assertEqual(instance.cast_time_zone_field, pytz.timezone('US/Eastern'))

        # Save and verify it is still the same
        instance.save()
        self.assertEqual(instance.cast_time_zone_field, pytz.timezone('US/Eastern'))

        # Load from db and verify again
        instance.refresh_from_db()
        self.assertEqual(instance.cast_time_zone_field, pytz.timezone('US/Eastern'))

    def test_no_cast_on_assign(self):
        """
        Verifies the base time zone field is not cast when being assigned
        """
        instance = FakeModel()
        instance.no_cast_time_zone_field = 'US/Eastern'

        # Verify the timezone is a string
        self.assertEqual(instance.no_cast_time_zone_field, 'US/Eastern')

        # Save and verify it is still a string
        instance.save()
        self.assertEqual(instance.no_cast_time_zone_field, 'US/Eastern')

        # Load from db and verify it is now cast as a time zone
        instance.refresh_from_db()
        self.assertEqual(instance.no_cast_time_zone_field, pytz.timezone('US/Eastern'))

    def test_all_time_zones_choices(self):
        """
        Verifies that all time zones are available in a method for usage in form choices
        """
        # Obtain a timezone that is in pytz.all_timezones, but not in pytz.common_timezones
        timezones = set(pytz.all_timezones) - set(pytz.common_timezones)
        timezone = timezones.pop()

        choices = {
            choice[0]
            for choice in TimeZoneField.get_all_choices()
        }

        self.assertTrue(timezone in choices)

    def test_common_time_zones_choices(self):
        """
        Verifies that all time zones are available in a method for usage in form choices
        """
        # Obtain a timezone that is in pytz.all_timezones, but not in pytz.common_timezones
        timezones = set(pytz.all_timezones) - set(pytz.common_timezones)
        timezone = timezones.pop()

        choices = {
            choice[0]
            for choice in TimeZoneField.get_common_choices()
        }
        self.assertTrue(timezone not in choices)



import functools

from django.conf import settings
from django.db import transaction, utils


def durable(func):
    """
    https://seddonym.me/2020/11/19/trouble-atomic/
    Decorator to ensure that a function is not being called within an atomic block.
    Usage:
        @durable
        def some_function(...):
            with transaction.atomic():
               ...
    Code decorated like this is guaranteed to be *durable* - that is, not at risk of being rolled
    back due to an exception that happens after this function has completed.
    This is achieved by enforcing that:
    1. The function does not begin in the context of a currently open transaction.
    2. The function does not leave work uncommitted.
    Warning: This may not work with SQLite (which requires workarounds for bugs in the stdlib
    sqlite3 module).
    Disabling this behaviour in tests
    ---------------------------------
    This behaviour doesn't play well with tests that are wrapped in transactions.
    It can be disabled by setting DISABLE_DURABILITY_CHECKING to True in your Django settings.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        durability_checking_enabled = not getattr(
            settings, "DISABLE_DURABILITY_CHECKING", False
        )
        if durability_checking_enabled and _is_in_atomic_block():
            raise utils.ProgrammingError(
                "A durable function must not be called within a database transaction."
            )

        return_value = func(*args, **kwargs)

        if durability_checking_enabled and _db_may_have_uncommitted_work():
            # Clean up first, otherwise we may see errors later that will mask this one.
            transaction.rollback()
            raise utils.ProgrammingError(
                "A durable function must not leave work uncommitted."
            )

        return return_value

    return wrapper


def _is_in_atomic_block():
    return not transaction.get_autocommit()


def _db_may_have_uncommitted_work():
    # Django doesn't seem to provide an API to tell this, but practically speaking there shouldn't
    # be any uncommitted work if we're not in an atomic block, and autocommit is turned on.
    return not transaction.get_autocommit()



# Generated by Django 2.1.7 on 2019-04-22 20:37

from django.db import migrations
import ambition_utils.fields


class Migration(migrations.Migration):

    dependencies = [
        ('rrule', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rrule',
            name='time_zone',
            field=ambition_utils.fields.TimeZoneField(default='UTC'),
        ),
    ]
